# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional, Union

import numpy as np
import torch
from torch.nn import Module

import mlrun
from mlrun.artifacts import Artifact
from mlrun.features import Feature

from .._dl_common import DLModelHandler
from .utils import PyTorchUtils


class PyTorchModelHandler(DLModelHandler):
    """
    Class for handling a PyTorch model, enabling loading and saving it during runs.
    """

    # Framework name:
    FRAMEWORK_NAME = "torch"

    # Declare a type of input sample:
    IOSample = Union[torch.Tensor, np.ndarray]

    class _LabelKeys:
        """
        Required labels keys to log with the model.
        """

        MODEL_CLASS_NAME = "model-class-name"

    def __init__(
        self,
        model: Module = None,
        model_path: Optional[str] = None,
        model_name: Optional[str] = None,
        model_class: Optional[Union[type[Module], str]] = None,
        modules_map: Optional[
            Union[dict[str, Union[None, str, list[str]]], str]
        ] = None,
        custom_objects_map: Optional[
            Union[dict[str, Union[str, list[str]]], str]
        ] = None,
        custom_objects_directory: Optional[str] = None,
        context: mlrun.MLClientCtx = None,
        **kwargs,
    ):
        """
        Initialize the handler. The model can be set here so it won't require loading.
        :param model:                    Model to handle or None in case a loading parameters were supplied.
        :param model_path:               Path to the model's directory with the saved '.pt' file. The file must start
                                         with the given model name. The model path can be also passed as a model object
                                         path in the following format:
                                         'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'.
        :param model_name:               The model name for saving and logging the model:
                                         * Mandatory for loading the model from a local path.
                                         * If given a logged model (store model path) it will be read from the artifact.
                                         * If given a loaded model object and the model name is None, the name will be
                                           set to the model's object name / class.
        :param model_class:              The model's class type object. Can be passed as the class's name (string) as
                                         well. The model class must appear in the custom objects / modules map
                                         dictionary / json. If the model path given is of a store object, this model
                                         class name will be read from the logged label of the model.
        :param modules_map:              A dictionary of all the modules required for loading the model. Each key
                                         is a path to a module and its value is the object name to import from it. All
                                         the modules will be imported globally. If multiple objects needed to be
                                         imported from the same module a list can be given. The map can be passed as a
                                         path to a json file as well. For example:
                                         {
                                             "module1": None,  # => import module1
                                             "module2": ["func1", "func2"],  # => from module2 import func1, func2
                                             "module3.sub_module": "func3",  # => from module3.sub_module import func3
                                         }
                                         If the model path given is of a store object, the modules map will be read from
                                         the logged modules map artifact of the model.
        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:
                                         {
                                             "/.../custom_optimizer.py": "optimizer",
                                             "/.../custom_layers.py": ["layer1", "layer2"]
                                         }
                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         If the model path given is of a store object, the custom objects map will be
                                         read from the logged custom object map artifact of the model.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model). If the model path given is of a store object, the
                                         custom objects files will be read from the logged custom object artifact of the
                                         model.
        :param context:                  MLRun context to work with for logging the model.

        :raise MLRunInvalidArgumentError: If the provided model path is of a local model files but the model class name
                                          and or the model name were not provided (= None).
        """
        # Validate a modules map or custom objects were provided:
        if (
            model_path is None
            and modules_map is None
            and (custom_objects_directory is None or custom_objects_directory is None)
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "At least 'modules_map' or both custom objects parameters: 'custom_objects_map' and "
                "'custom_objects_directory' are mandatory for the handler as the class must be located in a "
                "custom object python file or an installed module. Without one of them the model will not be able "
                "to be saved and logged"
            )

        # If the model is given try to set the model class:
        if model is not None:
            if model_class is None:
                model_class = type(model).__name__

        # Parse the class name (in case it was passed as a class type) and store it:
        if model_class is not None:
            self._model_class_name = (
                model_class if isinstance(model_class, str) else model_class.__name__
            )

        # Set up the base handler class:
        super().__init__(
            model=model,
            model_path=model_path,
            model_name=model_name,
            modules_map=modules_map,
            custom_objects_map=custom_objects_map,
            custom_objects_directory=custom_objects_directory,
            context=context,
            **kwargs,
        )

        # Set the required labels:
        self.set_labels()

    def set_labels(
        self,
        to_add: Optional[dict[str, Union[str, int, float]]] = None,
        to_remove: Optional[list[str]] = None,
    ):
        """
        Update the labels dictionary of this model artifact. There are required labels that cannot be edited or removed.

        :param to_add:    The labels to add.
        :param to_remove: A list of labels keys to remove.
        """
        # Update the user's labels:
        super().set_labels(to_add=to_add, to_remove=to_remove)

        # Set the required labels:
        self._labels[self._LabelKeys.MODEL_CLASS_NAME] = self._model_class_name

    def save(
        self, output_path: Optional[str] = None, **kwargs
    ) -> Union[dict[str, Artifact], None]:
        """
        Save the handled model at the given output path.

        :param output_path: The full path to the directory to save the handled model at. If not given, the context
                            stored will be used to save the model in the default location.

        :return The saved model additional artifacts (if needed) dictionary if context is available and None otherwise.

        :raise MLRunRuntimeError:         In case there is no model initialized in this handler.
        :raise MLRunInvalidArgumentError: If an output path was not given, yet a context was not provided in
                                          initialization.
        """
        super().save(output_path=output_path)

        # Set the output path:
        if output_path is None:
            output_path = os.path.join(self._context.artifact_path, self._model_name)

        # Save the model:
        self._model_file = f"{self._model_name}.pt"
        torch.save(self._model.state_dict(), self._model_file)

        return None

    def load(self, checkpoint: Optional[str] = None, **kwargs):
        """
        Load the specified model in this handler. If a checkpoint is required to be loaded, it can be given here
        according to the provided model path in the initialization of this handler. Additional parameters for the class
        initializer can be passed via the args list and kwargs dictionary.

        :param checkpoint: The checkpoint label to load the weights from. If the model path is of a store object, the
                           checkpoint will be taken from the logged checkpoints artifacts logged with the model. If the
                           model path is of a local directory, the checkpoint will be searched in it by the provided
                           name to this parameter.

        :raise MLRunInvalidArgumentError: If the model's class is not in the custom objects map.
        """
        super().load()

        # Validate the model's class is in the custom objects map:
        if (
            self._model_class_name not in self._custom_objects
            and self._model_class_name not in self._modules
        ):
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"The model class '{self._model_class_name}' was not found in the given custom objects map. The custom "
                f"objects map must include the model's class name in its values. Usually the model class should appear "
                f"last in the map dictionary as it is imported from the top to the bottom."
            )

        # Initialize the model:
        self._model = (
            self._custom_objects[self._model_class_name](**kwargs)
            if self._model_class_name in self._custom_objects
            else self._modules[self._model_class_name](**kwargs)
        )

        # Load the state dictionary into it:
        self._model.load_state_dict(torch.load(self._model_file))

    def to_onnx(
        self,
        model_name: Optional[str] = None,
        input_sample: Union[torch.Tensor, tuple[torch.Tensor, ...]] = None,
        input_layers_names: Optional[list[str]] = None,
        output_layers_names: Optional[list[str]] = None,
        dynamic_axes: Optional[dict[str, dict[int, str]]] = None,
        is_batched: bool = True,
        optimize: bool = True,
        output_path: Optional[str] = None,
        log: Optional[bool] = None,
    ):
        """
        Convert the model in this handler to an ONNX model. The layer names are optional, they do not change the
        semantics of the model, it is only for readability.

        :param model_name:          The name to give to the converted ONNX model. If not given the default name will be
                                    the stored model name with the suffix '_onnx'.
        :param optimize:            Whether to optimize the ONNX model using 'onnxoptimizer' before saving the model.
                                    Default: True.
        :param input_sample:        A torch.Tensor with the shape and data type of the expected input to the model. Can
                                    be passed as a tuple if the model expects multiple input tensors.
        :param input_layers_names:  List of names to assign to the input nodes of the graph in order. All of the other
                                    parameters (inner layers) can be set as well by passing additional names in the
                                    list. The order is by the order of the parameters in the model. If None, the inputs
                                    will be read from the handler's inputs. If it's also None, the default is:
                                    "input_0", "input_1", ...
        :param output_layers_names: List of names to assign to the output nodes of the graph in order. If None, the
                                    outputs will be read from the handler's outputs. If it's also None, the default
                                    is: "output_0" (for multiple outputs, this parameter must be provided).
        :param dynamic_axes:        If part of the input / output shape is dynamic, like (batch_size, 3, 32, 32) you can
                                    specify it by giving a dynamic axis to the input / output layer by its name as
                                    follows: {
                                        "input layer name": {0: "batch_size"},
                                        "output layer name": {0: "batch_size"},
                                    }
                                    If provided, the 'is_batched' flag will be ignored. Default: None.
        :param is_batched:          Whether to include a batch size as the first axis in every input and output layer.
                                    Default: True. Will be ignored if 'dynamic_axes' is provided.
        :param output_path:         In order to save the ONNX model, pass here the output directory. The model file will
                                    be named with the model name given. Default: None (not saving).
        :param log:                 In order to log the ONNX model, pass True. If None, the model will be logged if this
                                    handler has a MLRun context set. Default: None.

        :return: The converted ONNX model (onnx.ModelProto).

        :raise MLRunMissingDependencyError: If some of the ONNX packages are missing.
        """
        # Import onnx related modules:
        try:
            from mlrun.frameworks.onnx import ONNXModelHandler
        except ModuleNotFoundError:
            raise mlrun.errors.MLRunMissingDependencyError(
                "ONNX conversion requires additional packages to be installed. "
                "Please run 'pip install mlrun[pytorch]' to install MLRun's PyTorch package."
            )

        # Set the onnx model name:
        model_name = self._get_default_onnx_model_name(model_name=model_name)

        # Read the input signature in case its None:
        if input_sample is None:
            # Validate there are inputs set:
            if self._inputs is None:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "An input sample must be provided. Either use the method 'set_inputs' or pass an inputs sample "
                    "here."
                )
            # Parse the input features into a sample:
            input_sample = tuple(
                [
                    torch.zeros(
                        size=input_feature.dims,
                        dtype=PyTorchUtils.convert_value_type_to_torch_dtype(
                            value_type=input_feature.value_type
                        ),
                    )
                    for input_feature in self._inputs
                ]
            )
            if len(input_sample) == 1:
                input_sample = input_sample[0]

        # Set the default input layers names if not provided:
        if input_layers_names is None:
            input_layers_names = (
                (
                    [f"input_{i}" for i in range(len(input_sample))]
                    if isinstance(input_sample, tuple)
                    else ["input_0"]
                )
                if self._inputs is None
                else (
                    [
                        f"input_{i}" if layer.name == "" else layer.name
                        for i, layer in enumerate(self._inputs)
                    ]
                )
            )

        # Set the default output layers names if not provided:
        if output_layers_names is None:
            output_layers_names = (
                ["output_0"]
                if self._outputs is None
                else [
                    f"output_{i}" if layer.name == "" else layer.name
                    for i, layer in enumerate(self._outputs)
                ]
            )

        # Setup first axis to be a batch_size if needed:
        if dynamic_axes is None and is_batched:
            dynamic_axes = {
                layer: {0: "batch_size"}
                for layer in input_layers_names + output_layers_names
            }

        # Set the output model file:
        onnx_file = f"{model_name}.onnx"
        if output_path is None:
            output_path = "./"
        else:
            onnx_file = os.path.join(output_path, f"{model_name}.onnx")

        # Set the logging flag:
        log = self._context is not None if log is None else log

        # Convert to ONNX:
        torch.onnx.export(
            self._model,
            input_sample,
            onnx_file,
            input_names=input_layers_names,
            output_names=output_layers_names,
            dynamic_axes=dynamic_axes,
        )

        # Create a handler for the ONNX model:
        onnx_handler = ONNXModelHandler(
            model_name=model_name, model_path=output_path, context=self._context
        )

        # Pass on the inputs and outputs properties:
        if self._inputs is not None:
            onnx_handler.set_inputs(features=self._inputs)
        if self._outputs is not None:
            onnx_handler.set_outputs(features=self._outputs)

        # Load the ONNX model:
        onnx_handler.load()

        # Optimize the model if needed:
        if optimize:
            onnx_handler.optimize()
            # Save if logging is not required, as logging will save as well:
            if not log and output_path is not None:
                onnx_handler.save(output_path=output_path)

        # Log as a model object if needed:
        if log:
            onnx_handler.log()

        return onnx_handler.model

    def _collect_files_from_store_object(self):
        """
        If the model path given is of a store object, collect the needed model files into this handler for later loading
        the model.
        """
        # Read the model's class name:
        self._model_class_name = self._model_artifact.labels[
            self._LabelKeys.MODEL_CLASS_NAME
        ]

        # Continue collecting from abstract class:
        super()._collect_files_from_store_object()

    def _collect_files_from_local_path(self):
        """
        If the model path given is of a local path, search for the needed model files and collect them into this handler
        for later loading the model.

        :raise MLRunInvalidArgumentError: If the provided model class name from the user was None.
        :raise MLRunNotFoundError:        If the weights '.pt' file was not found.
        """
        # Read the model class provided:
        if self._model_class_name is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "The model class name must be provided when loading the model from local path. Otherwise, the handler "
                "will not be able to load the model."
            )

        # Collect the weights file:
        self._model_file = os.path.join(self._model_path, f"{self._model_name}.pt")
        if not os.path.exists(self._model_file):
            raise mlrun.errors.MLRunNotFoundError(
                f"The model weights file '{self._model_name}.pt' was not found within the given 'model_path': "
                f"'{self._model_path}'"
            )

    def _read_sample(self, sample: IOSample) -> Feature:
        """
        Read the sample into a MLRun Feature.

        :param sample: The sample to read.

        :return: The created Feature.

        :raise MLRunInvalidArgumentError: In case the given sample type cannot be read.
        """
        # Supported types:
        if isinstance(sample, np.ndarray):
            return super()._read_sample(sample=sample)
        elif isinstance(sample, torch.Tensor):
            return Feature(
                value_type=PyTorchUtils.convert_torch_dtype_to_value_type(
                    torch_dtype=sample.dtype
                ),
                dims=list(sample.shape),
            )

        # Unsupported type:
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"The sample type given '{type(sample)}' is not supported. The input / output ports are readable from "
            f"samples of the following types: torch.Tensor, np.ndarray."
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import numpy as np
import pandas as pd
from plotly.figure_factory import create_annotated_heatmap
from sklearn.metrics import confusion_matrix

from mlrun.artifacts import Artifact, PlotlyArtifact

from ..plan import MLPlanStages, MLPlotPlan
from ..utils import MLTypes, MLUtils


class ConfusionMatrixPlan(MLPlotPlan):
    """
    Plan for producing a confusion matrix.
    """

    _ARTIFACT_NAME = "confusion-matrix"

    def __init__(
        self,
        labels: np.ndarray = None,
        sample_weight: np.ndarray = None,
        normalize: Optional[str] = None,
    ):
        """
        Initialize a confusion matrix plan with the given configuration.

        To read more about the parameters, head to the SciKit-Learn docs at:
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.confusion_matrix.html

        :param labels:        Array of labels to index the matrix. This may be used to reorder or select a subset of the
                              labels.
        :param sample_weight: Sample weights to apply.
        :param normalize:     One of {'true', 'pred', 'all'} to normalize the confusion matrix over the true values
                              (rows), the predicted values (columns) or all. If None, confusion matrix will not be
                              normalized. Default: None.
        """
        # Store the parameters:
        self._labels = labels
        self._sample_weight = sample_weight
        self._normalize = normalize

        # Continue the initialization for the MLPlan:
        super().__init__()

    def is_ready(self, stage: MLPlanStages, is_probabilities: bool) -> bool:
        """
        Check whether the plan is fit for production by the given stage and prediction probabilities. The confusion
        matrix is ready only post prediction.

        :param stage:            The stage to check if the plan is ready.
        :param is_probabilities: True if the 'y_pred' that will be sent to 'produce' is a prediction of probabilities
                                 (from 'predict_proba') and False if not.

        :return: True if the plan is producible and False otherwise.
        """
        return stage == MLPlanStages.POST_PREDICT and not is_probabilities

    def produce(
        self,
        y: MLTypes.DatasetType,
        y_pred: MLTypes.DatasetType = None,
        model: MLTypes.ModelType = None,
        x: MLTypes.DatasetType = None,
        **kwargs,
    ) -> dict[str, Artifact]:
        """
        Produce the confusion matrix according to the ground truth (y) and predictions (y_pred) values. If predictions
        are not available, the model and a dataset can be given to produce them.

        :param y:      The ground truth values.
        :param y_pred: The predictions values.
        :param model:  Model to produce the predictions.
        :param x:      Input dataset to produce the predictions.

        :return: The produced confusion matrix artifact in an artifacts dictionary.
        """
        # Calculate the predictions if needed:
        y_pred = self._calculate_predictions(y_pred=y_pred, model=model, x=x)

        # Convert to DataFrame:
        y = MLUtils.to_dataframe(dataset=y)
        y_pred = MLUtils.to_dataframe(dataset=y_pred)

        # Set the labels array it not set:
        if self._labels is None:
            labels = pd.concat((y[y.columns[0]], y_pred[y_pred.columns[0]]))
            self._labels = np.sort(labels.unique()).tolist()

        # Calculate the confusion matrix:
        cm = confusion_matrix(
            y,
            y_pred,
            labels=self._labels,
            sample_weight=self._sample_weight,
            normalize=self._normalize,
        )

        # Initialize a plotly figure according to the targets (classes):
        figure = create_annotated_heatmap(
            cm,
            x=self._labels,
            y=self._labels,
            annotation_text=cm.astype(str),
            colorscale="Blues",
        )

        # Add title:
        figure.update_layout(
            title_text="Confusion matrix",
        )

        # Add custom x-axis title:
        figure.add_annotation(
            dict(
                font=dict(color="black", size=14),
                x=0.5,
                y=-0.1,
                showarrow=False,
                text="Predicted value",
                xref="paper",
                yref="paper",
            )
        )
        figure.update_xaxes(showline=True, linewidth=1, linecolor="black", mirror=True)

        # Add custom y-axis title:
        figure.add_annotation(
            dict(
                font=dict(color="black", size=14),
                x=-0.2,
                y=0.5,
                showarrow=False,
                text="Real value",
                textangle=-90,
                xref="paper",
                yref="paper",
            )
        )
        figure.update_yaxes(showline=True, linewidth=1, linecolor="black", mirror=True)

        # Adjust margins to make room for the y-axis title:
        figure.update_layout(margin=dict(t=100, l=100), width=500, height=500)

        # Add a color bar:
        figure["data"][0]["showscale"] = True
        figure["layout"]["yaxis"]["autorange"] = "reversed"

        # Create the plot's artifact:
        self._artifacts[self._ARTIFACT_NAME] = PlotlyArtifact(
            key=self._ARTIFACT_NAME,
            figure=figure,
        )

        return self._artifacts



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

from mlrun.errors import MLRunInvalidArgumentError


class LogHintKey:
    """
    Known keys for a log hint to have.
    """

    KEY = "key"
    ARTIFACT_TYPE = "artifact_type"
    EXTRA_DATA = "extra_data"
    METRICS = "metrics"


class LogHintUtils:
    """
    Static class for utilities functions to process log hints.
    """

    @staticmethod
    def parse_log_hint(
        log_hint: typing.Union[dict[str, str], str, None],
    ) -> typing.Union[dict[str, str], None]:
        """
        Parse a given log hint from string to a logging configuration dictionary. The string will be read as the
        artifact key ('key' in the dictionary) and if the string have a single colon, the following structure is
        assumed: "<artifact_key> : <artifact_type>".

        If a logging configuration dictionary is received, it will be validated to have a key field.

        None will be returned as None.

        :param log_hint: The log hint to parse.

        :return: The hinted logging configuration.

        :raise MLRunInvalidArgumentError: In case the log hint is not following the string structure or the dictionary
                                          is missing the key field.
        """
        # Check for None value:
        if log_hint is None:
            return None

        # If the log hint was provided as a string, construct a dictionary out of it:
        if isinstance(log_hint, str):
            # Check if only key is given:
            if ":" not in log_hint:
                log_hint = {LogHintKey.KEY: log_hint}
            # Check for valid "<key> : <artifact type>" pattern:
            else:
                if log_hint.count(":") > 1:
                    raise MLRunInvalidArgumentError(
                        f"Incorrect log hint pattern. Log hints can have only a single ':' in them to specify the "
                        f"desired artifact type the returned value will be logged as: "
                        f"'<artifact_key> : <artifact_type>', but given: {log_hint}"
                    )
                # Split into key and type:
                key, artifact_type = log_hint.replace(" ", "").split(":")
                if artifact_type == "":
                    raise MLRunInvalidArgumentError(
                        f"Incorrect log hint pattern. The ':' in a log hint should specify the desired artifact type "
                        f"the returned value will be logged as in the following pattern: "
                        f"'<artifact_key> : <artifact_type>', but no artifact type was given: {log_hint}"
                    )
                log_hint = {
                    LogHintKey.KEY: key,
                    LogHintKey.ARTIFACT_TYPE: artifact_type,
                }

        # Validate the log hint dictionary has the mandatory key:
        if LogHintKey.KEY not in log_hint:
            raise MLRunInvalidArgumentError(
                f"A log hint dictionary must include the 'key' - the artifact key (it's name). The following log hint "
                f"is missing the key: {log_hint}."
            )

        return log_hint



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def parse_versioned_object_uri(
    uri: str, default_project: str = ""
) -> tuple[str, str, str, str]:
    project = default_project
    tag = ""
    hash_key = ""
    if "/" in uri:
        loc = uri.find("/")
        project = uri[:loc]
        uri = uri[loc + 1 :]
    if ":" in uri:
        loc = uri.find(":")
        tag = uri[loc + 1 :]
        uri = uri[:loc]
    if "@" in uri:
        loc = uri.find("@")
        hash_key = uri[loc + 1 :]
        uri = uri[:loc]

    return project, uri, tag, hash_key


def generate_api_gateway_name(project: str, name: str) -> str:
    """
    Generate a unique (within project) api gateway name
    :param project: project name
    :param name: api gateway name

    :return: the resolved api gateway name
    """
    return f"{project}-{name}" if project else name



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import ClassVar, Literal, Optional, Union

import pandas as pd
import pydantic.v1
import v3io_frames.client

import mlrun.common.schemas.model_monitoring as mm_schemas
import mlrun.model_monitoring.db.tsdb.helpers
import mlrun.model_monitoring.helpers
from mlrun.utils import logger


class TSDBConnector(ABC):
    type: ClassVar[str]

    def __init__(self, project: str) -> None:
        """
        Initialize a new TSDB connector. The connector is used to interact with the TSDB and store monitoring data.
        At the moment we have 3 different types of monitoring data:
        - real time performance metrics: real time performance metrics that are being calculated by the model
        monitoring stream pod.
        Among these metrics are the base metrics (average latency and predictions over time), endpoint features
        (data samples), and custom metrics (user-defined metrics).
        - app_results: a detailed results that include status, kind, extra data, etc. These results are being calculated
        through the monitoring applications and stored in the TSDB using the model monitoring writer.
        - metrics: a basic key value that represents a numeric metric. Similar to the app_results, these metrics
        are being calculated through the monitoring applications and stored in the TSDB using the model monitoring
        writer.

        :param project: the name of the project.
        """
        self.project = project

    @abstractmethod
    def apply_monitoring_stream_steps(self, graph, **kwargs) -> None:
        """
        Apply TSDB steps on the provided monitoring graph. Throughout these steps, the graph stores live data of
        different key metric dictionaries. This data is being used by the monitoring dashboards in
        grafana.
        There are 3 different key metric dictionaries that are being generated throughout these steps:
        - base_metrics (average latency and predictions over time)
        - endpoint_features (Prediction and feature names and values)
        - custom_metrics (user-defined metrics)
        """
        pass

    def apply_writer_steps(self, graph, after, **kwargs) -> None:
        """
        Apply TSDB steps on the provided writer graph. Throughout these steps, the graph stores metrics / results.
        This data is being used by mlrun UI and the monitoring dashboards in grafana.
        There are 2 different key metric dictionaries that are being generated throughout these steps:
        - metrics (user-defined metrics) - model monitoring application metrics
        - results (user-defined results) - model monitoring application results
        """
        pass

    @abstractmethod
    def handle_model_error(self, graph, **kwargs) -> None:
        """
        Adds a branch to the stream pod graph to handle events that
        arrive with errors from the model server and saves them to the error TSDB table.
        The first step that generates by this method should come after `ForwardError` step.
        """

    @abstractmethod
    def write_application_event(
        self,
        event: dict,
        kind: mm_schemas.WriterEventKind = mm_schemas.WriterEventKind.RESULT,
    ) -> None:
        """
        Write a single application or metric to TSDB.

        :raise mlrun.errors.MLRunRuntimeError: If an error occurred while writing the event.
        """

    @abstractmethod
    def get_drift_data(
        self,
        start: datetime,
        end: datetime,
    ) -> mm_schemas.ModelEndpointDriftValues:
        """
        Fetches drift counts per interval in the specified time range.

        :param start: The start time of the query.
        :param end:   The end time of the query.

        :return: A ModelEndpointDriftValues object containing drift data.
        """

    @abstractmethod
    def delete_tsdb_records(self, endpoint_ids: list[str]) -> None:
        """
        Delete model endpoint records from the TSDB connector.

        :param endpoint_ids: List of model endpoint unique identifiers.
        """
        pass

    @abstractmethod
    def delete_application_records(
        self, application_name: str, endpoint_ids: Optional[list[str]] = None
    ) -> None:
        """
        Delete application records from the TSDB for the given model endpoints or all if ``None``.

        :param application_name: The name of the application to delete records for.
        :param endpoint_ids:     List of model endpoint unique identifiers.
        """
        pass

    @abstractmethod
    def delete_tsdb_resources(self):
        """
        Delete all project resources in the TSDB connector, such as model endpoints data and drift results.
        """
        pass

    @abstractmethod
    def get_model_endpoint_real_time_metrics(
        self,
        endpoint_id: str,
        metrics: list[str],
        start: str,
        end: str,
    ) -> dict[str, list[tuple[str, float]]]:
        """
        Getting real time metrics from the TSDB. There are pre-defined metrics for model endpoints such as
        `predictions_per_second` and `latency_avg_5m` but also custom metrics defined by the user. Note that these
        metrics are being calculated by the model monitoring stream pod.
        :param endpoint_id:      The unique id of the model endpoint.
        :param metrics:          A list of real-time metrics to return for the model endpoint.
        :param start:            The start time of the metrics. Can be represented by a string containing an  RFC 3339
                                 time, a  Unix timestamp in milliseconds, a relative time (`'now'` or
                                 `'now-[0-9]+[mhd]'`, where `m` = minutes, `h` = hours, `'d'` = days, and `'s'`
                                 = seconds), or 0 for the earliest time.
        :param end:              The end time of the metrics. Can be represented by a string containing an  RFC 3339
                                 time, a  Unix timestamp in milliseconds, a relative time (`'now'` or
                                 `'now-[0-9]+[mhd]'`, where `m` = minutes, `h` = hours, `'d'` = days, and `'s'`
                                 = seconds), or 0 for the earliest time.
        :return: A dictionary of metrics in which the key is a metric name and the value is a list of tuples that
                 includes timestamps and the values.
        """
        pass

    @abstractmethod
    def create_tables(self) -> None:
        """
        Create the TSDB tables using the TSDB connector. At the moment we support 3 types of tables:
        - app_results: a detailed result that includes status, kind, extra data, etc.
        - metrics: a basic key value that represents a numeric metric.
        - predictions: latency of each prediction.
        """

    @abstractmethod
    def read_metrics_data(
        self,
        *,
        endpoint_id: str,
        start: datetime,
        end: datetime,
        metrics: list[mm_schemas.ModelEndpointMonitoringMetric],
        type: Literal["metrics", "results"],
        with_result_extra_data: bool,
    ) -> Union[
        list[
            Union[
                mm_schemas.ModelEndpointMonitoringResultValues,
                mm_schemas.ModelEndpointMonitoringMetricNoData,
            ],
        ],
        list[
            Union[
                mm_schemas.ModelEndpointMonitoringMetricValues,
                mm_schemas.ModelEndpointMonitoringMetricNoData,
            ],
        ],
    ]:
        """
        Read metrics OR results from the TSDB and return as a list.

        :param endpoint_id: The model endpoint identifier.
        :param start:                  The start time of the query.
        :param end:                    The end time of the query.
        :param metrics:                The list of metrics to get the values for.
        :param type:                   "metrics" or "results" - the type of each item in metrics.
        :param with_result_extra_data: Whether to include the extra data in the results, relevant only when
                                       `type="results"`.
        :return:                        A list of result values or a list of metric values.
        """

    @abstractmethod
    def read_predictions(
        self,
        *,
        endpoint_id: str,
        start: datetime,
        end: datetime,
        aggregation_window: Optional[str] = None,
        agg_funcs: Optional[list[str]] = None,
        limit: Optional[int] = None,
    ) -> Union[
        mm_schemas.ModelEndpointMonitoringMetricValues,
        mm_schemas.ModelEndpointMonitoringMetricNoData,
    ]:
        """
        Read the "invocations" metric for the provided model endpoint in the given time range,
        and return the metric values if any, otherwise signify with the "no data" object.

        :param endpoint_id:        The model endpoint identifier.
        :param start:              The start time of the query.
        :param end:                The end time of the query.
        :param aggregation_window: On what time window length should the invocations be aggregated. If provided,
                                   the `agg_funcs` must be provided as well. Provided as a string in the format of '1m',
                                   '1h', etc.
        :param agg_funcs:          List of aggregation functions to apply on the invocations. If provided, the
                                   `aggregation_window` must be provided as well. Provided as a list of strings in
                                   the format of ['sum', 'avg', 'count', ...]
        :param limit:              The maximum number of records to return.

        :raise mlrun.errors.MLRunInvalidArgumentError: If only one of `aggregation_window` and `agg_funcs` is provided.
        :return:                   Metric values object or no data object.
        """

    @abstractmethod
    def get_last_request(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> Union[pd.DataFrame, dict[str, float]]:
        """
        Fetches data from the predictions TSDB table and returns the most recent request
        timestamp for each specified endpoint.

        :param endpoint_ids:    A list of model endpoint identifiers.
        :param start:           The start time for the query.
        :param end:             The end time for the query.

        :return: A pd.DataFrame containing the columns [endpoint_id, last_request, last_latency] or a dictionary
        containing the endpoint_id as the key and the last request timestamp as the value.
        if an endpoint has not been invoked within the specified time range, it will not appear in the result (relevant
        only to non-v3io connector).
        """

    @abstractmethod
    def get_drift_status(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> Union[pd.DataFrame, list[v3io_frames.client.RawFrame]]:
        """
        Fetches data from the app-results TSDB table and returns the highest status among all
        the result in the provided time range, which by default is the last 24 hours, for each specified endpoint.

        :param endpoint_ids:    A list of model endpoint identifiers.
        :param start:           The start time for the query.
        :param end:             The end time for the query.
        :param get_raw:         Whether to return the request as raw frames rather than a pandas dataframe. Defaults
          to False. This can greatly improve performance when a dataframe isn't needed.

        :return: A pd.DataFrame containing the columns [result_status, endpoint_id].
        If an endpoint has not been monitored within the specified time range (last 24 hours),
        it will not appear in the result.
        """

    @abstractmethod
    def get_metrics_metadata(
        self,
        endpoint_id: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> pd.DataFrame:
        """
        Fetches distinct metrics metadata from the metrics TSDB table for a specified model endpoints.

        :param endpoint_id:        The model endpoint identifier. Can be a single id or a list of ids.
        :param start:              The start time of the query.
        :param end:                The end time of the query.

        :return: A pd.DataFrame containing all distinct metrics for the specified endpoint within the given time range.
        Containing the columns [application_name, metric_name, endpoint_id]
        """

    @abstractmethod
    def get_results_metadata(
        self,
        endpoint_id: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> pd.DataFrame:
        """
        Fetches distinct results metadata from the app-results TSDB table for a specified model endpoints.

        :param endpoint_id:        The model endpoint identifier. Can be a single id or a list of ids.
        :param start:              The start time of the query.
        :param end:                The end time of the query.

        :return: A pd.DataFrame containing all distinct results for the specified endpoint within the given time range.
        Containing the columns [application_name, result_name, result_kind, endpoint_id]
        """

    @abstractmethod
    def get_error_count(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> Union[pd.DataFrame, list[v3io_frames.client.RawFrame]]:
        """
        Fetches data from the error TSDB table and returns the error count for each specified endpoint.

        :param endpoint_ids:    A list of model endpoint identifiers.
        :param start:           The start time for the query.
        :param end:             The end time for the query.
        :param get_raw:         Whether to return the request as raw frames rather than a pandas dataframe. Defaults
          to False. This can greatly improve performance when a dataframe isn't needed.

        :return: A pd.DataFrame containing the columns [error_count, endpoint_id].
        If an endpoint have not raised error within the specified time range, it will not appear in the result.
        """

    @abstractmethod
    def get_avg_latency(
        self,
        endpoint_ids: Union[str, list[str]],
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        get_raw: bool = False,
    ) -> Union[pd.DataFrame, list[v3io_frames.client.RawFrame]]:
        """
        Fetches data from the predictions TSDB table and returns the average latency for each specified endpoint
        in the provided time range, which by default is the last 24 hours.

        :param endpoint_ids:    A list of model endpoint identifiers.
        :param start:           The start time for the query.
        :param end:             The end time for the query.
        :param get_raw:         Whether to return the request as raw frames rather than a pandas dataframe. Defaults
          to False. This can greatly improve performance when a dataframe isn't needed.

        :return: A pd.DataFrame containing the columns [avg_latency, endpoint_id].
        If an endpoint has not been invoked within the specified time range, it will not appear in the result.
        """

    @abstractmethod
    def count_results_by_status(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        endpoint_ids: Optional[Union[str, list[str]]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
        result_status_list: Optional[list[int]] = None,
    ) -> dict[tuple[str, int], int]:
        """
        Read results status from the TSDB and return a dictionary of results statuses by application name.

        :param start:              The start time in which to read the results. By default, the last 24 hours are read.
        :param end:                The end time in which to read the results. Default is the current time (now).
        :param endpoint_ids:       Optional list of endpoint ids to filter the results by. By default, all
                                   endpoint ids are included.
        :param application_names:  Optional list of application names to filter the results by. By default, all
                                   application are included.
        :param result_status_list: Optional list of result statuses to filter the results by. By default, all
                                   result statuses are included.

        :return: A dictionary where the key is a tuple of (application_name, result_status) and the value is the total
                 number of results with that status for that application.
                 For example:
                 {
                    ('app1', 1): 10,
                    ('app1', 2): 5
                 }
        """

    @abstractmethod
    def count_processed_model_endpoints(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
    ) -> dict[str, int]:
        """
        Count the number of processed model endpoints within a given time range for specific applications.
        :param start:              The start time of the query. Last 24 hours is used by default.
        :param end:                The end time of the query. The current time is used by default.
        :param application_names:  A list of application names to filter the results by. If not provided, all
                                   applications are included.
        :return:                   The count of processed model endpoints.
        """

    @abstractmethod
    def calculate_latest_metrics(
        self,
        start: Optional[Union[datetime, str]] = None,
        end: Optional[Union[datetime, str]] = None,
        application_names: Optional[Union[str, list[str]]] = None,
    ) -> list[
        Union[mm_schemas.ApplicationResultRecord, mm_schemas.ApplicationMetricRecord]
    ]:
        """
        Calculate the latest metrics and results across applications.
        :param start:              The start time of the query. Last 24 hours is used by default.
        :param end:                The end time of the query. The current time is used by default.
        :param application_names:  A list of application names to filter the results by. If not provided, all
                                   applications are included.
        :return:                   A list containing the latest metrics and results for each application.
                                   example::
                                   [
                                       {
                                           "type": "metric",
                                           "time": "2025-06-29 13:36:37 +00:00",
                                           "metric_name": "hellinger_mean",
                                           "value": 0.123456,
                                       },
                                        {
                                             "type": "result",
                                             "time": "2025-06-29 13:36:37 +00:00",
                                             "result_name": "drift_status",
                                             "kind": "2",
                                             "status": 0,
                                             "value": 15.4,
                                        },
                                       ...
                                   ]
        """

    def add_basic_metrics(
        self,
        model_endpoint_objects: list[mlrun.common.schemas.ModelEndpoint],
        metric_list: Optional[list[str]] = None,
    ) -> list[mlrun.common.schemas.ModelEndpoint]:
        raise NotImplementedError()

    @staticmethod
    def df_to_metrics_values(
        *,
        df: pd.DataFrame,
        metrics: list[mm_schemas.ModelEndpointMonitoringMetric],
        project: str,
    ) -> list[
        Union[
            mm_schemas.ModelEndpointMonitoringMetricValues,
            mm_schemas.ModelEndpointMonitoringMetricNoData,
        ]
    ]:
        """
        Parse a time-indexed DataFrame of metrics from the TSDB into a list of
        metrics values per distinct results.
        When a metric is not found in the DataFrame, it is represented in a no-data object.
        """
        metrics_without_data = {metric.full_name: metric for metric in metrics}

        metrics_values: list[
            Union[
                mm_schemas.ModelEndpointMonitoringMetricValues,
                mm_schemas.ModelEndpointMonitoringMetricNoData,
            ]
        ] = []
        if not df.empty:
            grouped = df.groupby(
                [
                    mm_schemas.WriterEvent.APPLICATION_NAME,
                    mm_schemas.MetricData.METRIC_NAME,
                ],
                observed=False,
            )
        else:
            logger.debug("No metrics", missing_metrics=metrics_without_data.keys())
            grouped = []
        for (app_name, name), sub_df in grouped:
            full_name = mm_schemas.model_endpoints.compose_full_name(
                project=project,
                app=app_name,
                name=name,
                type=mm_schemas.ModelEndpointMonitoringMetricType.METRIC,
            )
            metrics_values.append(
                mm_schemas.ModelEndpointMonitoringMetricValues(
                    full_name=full_name,
                    values=list(
                        zip(
                            sub_df.index,
                            sub_df[mm_schemas.MetricData.METRIC_VALUE],
                        )
                    ),  # pyright: ignore[reportArgumentType]
                )
            )
            del metrics_without_data[full_name]

        for metric in metrics_without_data.values():
            metrics_values.append(
                mm_schemas.ModelEndpointMonitoringMetricNoData(
                    full_name=metric.full_name,
                    type=mm_schemas.ModelEndpointMonitoringMetricType.METRIC,
                )
            )

        return metrics_values

    @staticmethod
    def df_to_results_values(
        *,
        df: pd.DataFrame,
        metrics: list[mm_schemas.ModelEndpointMonitoringMetric],
        project: str,
    ) -> list[
        Union[
            mm_schemas.ModelEndpointMonitoringResultValues,
            mm_schemas.ModelEndpointMonitoringMetricNoData,
        ]
    ]:
        """
        Parse a time-indexed DataFrame of results from the TSDB into a list of
        results values per distinct results.
        When a result is not found in the DataFrame, it is represented in no-data object.
        """
        metrics_without_data = {metric.full_name: metric for metric in metrics}

        metrics_values: list[
            Union[
                mm_schemas.ModelEndpointMonitoringResultValues,
                mm_schemas.ModelEndpointMonitoringMetricNoData,
            ]
        ] = []
        if not df.empty:
            grouped = df.groupby(
                [
                    mm_schemas.WriterEvent.APPLICATION_NAME,
                    mm_schemas.ResultData.RESULT_NAME,
                ],
                observed=False,
            )
        else:
            grouped = []
            logger.debug("No results", missing_results=metrics_without_data.keys())
        for (app_name, name), sub_df in grouped:
            result_kind = mlrun.model_monitoring.db.tsdb.helpers._get_result_kind(
                sub_df
            )
            full_name = mm_schemas.model_endpoints.compose_full_name(
                project=project, app=app_name, name=name
            )
            try:
                metrics_values.append(
                    mm_schemas.ModelEndpointMonitoringResultValues(
                        full_name=full_name,
                        result_kind=result_kind,
                        values=list(
                            zip(
                                sub_df.index,
                                sub_df[mm_schemas.ResultData.RESULT_VALUE],
                                sub_df[mm_schemas.ResultData.RESULT_STATUS],
                                sub_df[mm_schemas.ResultData.RESULT_EXTRA_DATA],
                            )
                        ),  # pyright: ignore[reportArgumentType]
                    )
                )
            except pydantic.v1.ValidationError:
                logger.exception(
                    "Failed to convert data-frame into `ModelEndpointMonitoringResultValues`",
                    full_name=full_name,
                    sub_df_json=sub_df.to_json(),
                )
                raise
            del metrics_without_data[full_name]

        for metric in metrics_without_data.values():
            if metric.full_name == mlrun.model_monitoring.helpers.get_invocations_fqn(
                project
            ):
                continue
            metrics_values.append(
                mm_schemas.ModelEndpointMonitoringMetricNoData(
                    full_name=metric.full_name,
                    type=mm_schemas.ModelEndpointMonitoringMetricType.RESULT,
                )
            )

        return metrics_values

    @staticmethod
    def df_to_metrics_list(
        *,
        df: pd.DataFrame,
        project: str,
        type: str,
    ) -> list[mm_schemas.ModelEndpointMonitoringMetric]:
        """
        Parse a DataFrame of metrics from the TSDB into a list of mm metrics objects.

        :param df:      The DataFrame to parse.
        :param project: The project name.
        :param type:    The type of the metrics (either "result" or "metric").

        :return:        A list of mm metrics objects.
        """

        return list(
            map(
                lambda record: mm_schemas.ModelEndpointMonitoringMetric(
                    project=project,
                    type=type,
                    app=record.get(mm_schemas.WriterEvent.APPLICATION_NAME),
                    name=record.get(mm_schemas.ResultData.RESULT_NAME)
                    or record.get(mm_schemas.MetricData.METRIC_NAME),
                    kind=record.get(mm_schemas.ResultData.RESULT_KIND),
                ),
                df.to_dict("records"),
            )
        )

    @staticmethod
    def df_to_metrics_grouped_dict(
        *,
        df: pd.DataFrame,
        project: str,
        type: str,
    ) -> dict[str, list[mm_schemas.ModelEndpointMonitoringMetric]]:
        """
        Parse a DataFrame of metrics from the TSDB into a grouped mm metrics objects by endpoint_id.

        :param df:      The DataFrame to parse.
        :param project: The project name.
        :param type:    The type of the metrics (either "result" or "metric").

        :return:        A grouped dict of mm metrics/results, using model_endpoints_ids as keys.
        """

        if df.empty:
            return {}

        grouped_by_fields = [mm_schemas.WriterEvent.APPLICATION_NAME]
        if type == "result":
            name_column = mm_schemas.ResultData.RESULT_NAME
            grouped_by_fields.append(mm_schemas.ResultData.RESULT_KIND)
        else:
            name_column = mm_schemas.MetricData.METRIC_NAME

        grouped_by_fields.append(name_column)
        # groupby has different behavior for category columns
        df["endpoint_id"] = df["endpoint_id"].astype(str)
        grouped_by_df = df.groupby("endpoint_id")
        grouped_dict = grouped_by_df.apply(
            lambda group: list(
                map(
                    lambda record: mm_schemas.ModelEndpointMonitoringMetric(
                        project=project,
                        type=type,
                        app=record.get(mm_schemas.WriterEvent.APPLICATION_NAME),
                        name=record.get(name_column),
                        **{"kind": record.get(mm_schemas.ResultData.RESULT_KIND)}
                        if type == "result"
                        else {},
                    ),
                    group[grouped_by_fields].to_dict(orient="records"),
                )
            )
        ).to_dict()
        return grouped_dict

    @staticmethod
    def df_to_events_intersection_dict(
        *,
        df: pd.DataFrame,
        project: str,
        type: Union[str, mm_schemas.ModelEndpointMonitoringMetricType],
    ) -> dict[str, list[mm_schemas.ModelEndpointMonitoringMetric]]:
        """
        Parse a DataFrame of metrics from the TSDB into a dict of intersection metrics/results by name and application
         (and kind in results).

        :param df:      The DataFrame to parse.
        :param project: The project name.
        :param type:    The type of the metrics (either "result" or "metric").

        :return:        A dictionary where the key is event type (as defined by `INTERSECT_DICT_KEYS`),
                        and the value is a list containing the intersect metrics or results across all endpoint IDs.

                        For example:
                        {
                            "intersect_metrics": [...]
                        }
        """
        dict_key = mm_schemas.INTERSECT_DICT_KEYS[type]
        metrics = []
        if df.empty:
            return {dict_key: []}

        columns_to_zip = [mm_schemas.WriterEvent.APPLICATION_NAME]

        if type == "result":
            name_column = mm_schemas.ResultData.RESULT_NAME
            columns_to_zip.append(mm_schemas.ResultData.RESULT_KIND)
        else:
            name_column = mm_schemas.MetricData.METRIC_NAME
        columns_to_zip.insert(1, name_column)

        # groupby has different behavior for category columns
        df["endpoint_id"] = df["endpoint_id"].astype(str)
        df["event_values"] = list(zip(*[df[col] for col in columns_to_zip]))
        grouped_by_event_values = df.groupby("endpoint_id")["event_values"].apply(set)
        common_event_values_combinations = set.intersection(*grouped_by_event_values)
        result_kind = None
        for data in common_event_values_combinations:
            application_name, event_name = data[0], data[1]
            if len(data) > 2:  # in result case
                result_kind = data[2]
            metrics.append(
                mm_schemas.ModelEndpointMonitoringMetric(
                    project=project,
                    type=type,
                    app=application_name,
                    name=event_name,
                    kind=result_kind,
                )
            )
        return {dict_key: metrics}

    @staticmethod
    def _prepare_aligned_start_end(
        start: datetime, end: datetime
    ) -> tuple[datetime, datetime, str]:
        delta = end - start
        if delta <= timedelta(hours=6):
            interval = "10m"
            start = start.replace(
                minute=start.minute // 10 * 10, second=0, microsecond=0
            )
        elif delta <= timedelta(hours=72):
            interval = "1h"
            start = start.replace(minute=0, second=0, microsecond=0)
        else:
            interval = "1d"
            start = start.replace(hour=0, minute=0, second=0, microsecond=0)

        interval_map = {
            "10m": timedelta(minutes=10),
            "1h": timedelta(hours=1),
            "1d": timedelta(days=1),
        }
        delta = end - start
        interval_td = interval_map[interval]
        end = start + (delta // interval_td) * interval_td
        return start, end, interval

    @staticmethod
    def _df_to_drift_data(df: pd.DataFrame) -> mm_schemas.ModelEndpointDriftValues:
        suspected_val = mm_schemas.constants.ResultStatusApp.potential_detection.value
        detected_val = mm_schemas.constants.ResultStatusApp.detected.value
        aggregated_df = (
            df.groupby(["_wstart", f"max({mm_schemas.ResultData.RESULT_STATUS})"])
            .size()  # add size column for each interval x result-status combination
            .unstack()  # create a size column for each result-status
            .reindex(
                columns=[suspected_val, detected_val], fill_value=0
            )  # ensure both columns exists
            .fillna(0)
            .astype(int)
            .rename(
                columns={
                    suspected_val: "count_suspected",
                    detected_val: "count_detected",
                }
            )
        )
        values = list(
            zip(
                aggregated_df.index,
                aggregated_df["count_suspected"],
                aggregated_df["count_detected"],
            )
        )
        return mm_schemas.ModelEndpointDriftValues(values=values)

    def add_pre_writer_steps(self, graph, after):
        return None



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from os import environ

from ..config import config
from .base import RunDBError, RunDBInterface  # noqa


def get_or_set_dburl(default=""):
    if not config.dbpath and default:
        config.dbpath = default
        environ["MLRUN_DBPATH"] = default
    return config.dbpath


def get_run_db(url="", secrets=None, force_reconnect=False) -> RunDBInterface:
    """Returns the runtime database"""
    # import here to avoid circular import
    import mlrun.db.factory

    run_db_factory = mlrun.db.factory.RunDBFactory()
    return run_db_factory.create_run_db(url, secrets, force_reconnect)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import typing
from datetime import datetime, timezone
from typing import Any, Callable, NewType, Optional

import storey

import mlrun.common.model_monitoring
import mlrun.common.schemas
import mlrun.common.schemas.alert as alert_objects
import mlrun.model_monitoring
from mlrun.common.schemas.model_monitoring.constants import (
    HistogramDataDriftApplicationConstants,
    MetricData,
    ResultData,
    ResultKindApp,
    ResultStatusApp,
    StatsData,
    StatsKind,
    WriterEvent,
    WriterEventKind,
)
from mlrun.config import config
from mlrun.model_monitoring.db import TSDBConnector
from mlrun.model_monitoring.db._stats import (
    ModelMonitoringCurrentStatsFile,
    ModelMonitoringDriftMeasuresFile,
)
from mlrun.model_monitoring.helpers import get_result_instance_fqn
from mlrun.serving.utils import StepToDict
from mlrun.utils import logger

_RawEvent = dict[str, Any]
_AppResultEvent = NewType("_AppResultEvent", _RawEvent)


class _WriterEventError:
    pass


class _WriterEventValueError(_WriterEventError, ValueError):
    pass


class _WriterEventTypeError(_WriterEventError, TypeError):
    pass


class ModelMonitoringWriter(StepToDict):
    """
    Write monitoring application results to the target databases
    """

    kind = "monitoring_application_stream_pusher"

    def __init__(
        self,
        project: str,
        secret_provider: Optional[Callable] = None,
    ) -> None:
        self.project = project
        self.name = project  # required for the deployment process

        self._tsdb_connector = mlrun.model_monitoring.get_tsdb_connector(
            project=self.project, secret_provider=secret_provider
        )

    def _generate_event_on_drift(
        self,
        entity_id: str,
        result_status: int,
        event_value: dict,
        project_name: str,
        result_kind: int,
    ) -> None:
        entity = mlrun.common.schemas.alert.EventEntities(
            kind=alert_objects.EventEntityKind.MODEL_ENDPOINT_RESULT,
            project=project_name,
            ids=[entity_id],
        )

        event_kind = self._generate_alert_event_kind(
            result_status=result_status, result_kind=result_kind
        )

        event_data = mlrun.common.schemas.Event(
            kind=alert_objects.EventKind(value=event_kind),
            entity=entity,
            value_dict=event_value,
        )
        logger.info("Sending a drift event")
        mlrun.get_run_db().generate_event(event_kind, event_data)
        logger.info("Drift event sent successfully")

    @staticmethod
    def _generate_alert_event_kind(
        result_kind: int, result_status: int
    ) -> alert_objects.EventKind:
        """Generate the required Event Kind format for the alerting system"""
        event_kind = ResultKindApp(value=result_kind).name

        if result_status == ResultStatusApp.detected.value:
            event_kind = f"{event_kind}_detected"
        else:
            event_kind = f"{event_kind}_suspected"
        return alert_objects.EventKind(
            value=mlrun.utils.helpers.normalize_name(event_kind)
        )

    @staticmethod
    def _reconstruct_event(event: _RawEvent) -> tuple[_AppResultEvent, WriterEventKind]:
        """
        Modify the raw event into the expected monitoring application event
        schema as defined in `mlrun.common.schemas.model_monitoring.constants.WriterEvent`
        """
        if not isinstance(event, dict):
            raise _WriterEventTypeError(
                f"The event is of type: {type(event)}, expected a dictionary"
            )
        kind = event.pop(WriterEvent.EVENT_KIND, WriterEventKind.RESULT)
        result_event = _AppResultEvent(json.loads(event.pop(WriterEvent.DATA, "{}")))
        result_event.update(_AppResultEvent(event))

        expected_keys = list(
            set(WriterEvent.list()).difference(
                [WriterEvent.EVENT_KIND, WriterEvent.DATA]
            )
        )
        if kind == WriterEventKind.METRIC:
            expected_keys.extend(MetricData.list())
        elif kind == WriterEventKind.RESULT:
            expected_keys.extend(ResultData.list())
        elif kind == WriterEventKind.STATS:
            expected_keys.extend(StatsData.list())
        else:
            raise _WriterEventValueError(
                f"Unknown event kind: {kind}, expected one of: {WriterEventKind.list()}"
            )
        missing_keys = [key for key in expected_keys if key not in result_event]
        if missing_keys:
            raise _WriterEventValueError(
                f"The received event misses some keys compared to the expected "
                f"monitoring application event schema: {missing_keys} for event kind {kind}"
            )

        return result_event, kind

    def write_stats(self, event: _AppResultEvent) -> None:
        """
        Write to file the application stats event
        :param event: application stats event
        """
        endpoint_id = event[WriterEvent.ENDPOINT_ID]
        logger.debug(
            "Updating the model endpoint with stats",
            endpoint_id=endpoint_id,
        )
        stat_kind = event.get(StatsData.STATS_NAME)
        data, timestamp_str = event.get(StatsData.STATS), event.get(StatsData.TIMESTAMP)
        timestamp = datetime.fromisoformat(timestamp_str).astimezone(tz=timezone.utc)
        if stat_kind == StatsKind.CURRENT_STATS.value:
            ModelMonitoringCurrentStatsFile(self.project, endpoint_id).write(
                data, timestamp
            )
        elif stat_kind == StatsKind.DRIFT_MEASURES.value:
            ModelMonitoringDriftMeasuresFile(self.project, endpoint_id).write(
                data, timestamp
            )
        logger.info(
            "Updated the model endpoint statistics",
            endpoint_id=endpoint_id,
            stats_kind=stat_kind,
        )

    def do(self, event: _RawEvent) -> None:
        event, kind = self._reconstruct_event(event)
        logger.info("Starting to write event", event=event)
        if (
            kind == WriterEventKind.STATS
            and event[WriterEvent.APPLICATION_NAME]
            == HistogramDataDriftApplicationConstants.NAME
        ):
            self.write_stats(event)
            logger.info("Model monitoring writer finished handling event")
            return
        self._tsdb_connector.write_application_event(event=event.copy(), kind=kind)

        logger.info("Completed event DB writes")

        if (
            mlrun.mlconf.alerts.mode == mlrun.common.schemas.alert.AlertsModes.enabled
            and kind == WriterEventKind.RESULT
            and (
                event[ResultData.RESULT_STATUS] == ResultStatusApp.detected.value
                or event[ResultData.RESULT_STATUS]
                == ResultStatusApp.potential_detection.value
            )
        ):
            event_value = {
                "app_name": event[WriterEvent.APPLICATION_NAME],
                "model": event[WriterEvent.ENDPOINT_NAME],
                "model_endpoint_id": event[WriterEvent.ENDPOINT_ID],
                "result_name": event[ResultData.RESULT_NAME],
                "result_value": event[ResultData.RESULT_VALUE],
            }
            self._generate_event_on_drift(
                entity_id=get_result_instance_fqn(
                    event[WriterEvent.ENDPOINT_ID],
                    event[WriterEvent.APPLICATION_NAME],
                    event[ResultData.RESULT_NAME],
                ),
                result_status=event[ResultData.RESULT_STATUS],
                event_value=event_value,
                project_name=self.project,
                result_kind=event[ResultData.RESULT_KIND],
            )

        logger.info("Model monitoring writer finished handling event")


class WriterGraphFactory:
    def __init__(
        self,
        parquet_path: str,
    ):
        self.parquet_path = parquet_path
        self.parquet_batching_max_events = (
            config.model_endpoint_monitoring.writer_graph.max_events
        )
        self.parquet_batching_timeout_secs = (
            config.model_endpoint_monitoring.writer_graph.parquet_batching_timeout_secs
        )

    def apply_writer_graph(
        self,
        fn: mlrun.runtimes.ServingRuntime,
        tsdb_connector: TSDBConnector,
    ):
        graph = typing.cast(
            mlrun.serving.states.RootFlowStep,
            fn.set_topology(mlrun.serving.states.StepKinds.flow, engine="async"),
        )

        graph.to("ReconstructWriterEvent", "event_reconstructor")
        step = tsdb_connector.add_pre_writer_steps(
            graph=graph, after="event_reconstructor"
        )
        before_choice = step.name if step else "event_reconstructor"
        graph.add_step("KindChoice", "kind_choice_step", after=before_choice)
        tsdb_connector.apply_writer_steps(
            graph=graph,
            after="kind_choice_step",
        )
        graph.add_step(
            "AlertGenerator",
            "alert_generator",
            after="kind_choice_step",
            project=fn.metadata.project,
        )
        graph.add_step(
            "storey.Filter",
            name="filter_none",
            _fn="(event is not None)",
            after="alert_generator",
        )
        graph.add_step(
            "mlrun.serving.remote.MLRunAPIRemoteStep",
            name="alert_generator_api_call",
            after="filter_none",
            method="POST",
            path=f"projects/{fn.metadata.project}/events/{{kind}}",
            fill_placeholders=True,
        )

        graph.add_step(
            "mlrun.datastore.storeytargets.ParquetStoreyTarget",
            alternative_v3io_access_key=mlrun.common.schemas.model_monitoring.ProjectSecretKeys.ACCESS_KEY,
            name="stats_writer",
            after="kind_choice_step",
            graph_shape="cylinder",
            path=self.parquet_path
            if self.parquet_path.endswith("/")
            else self.parquet_path + "/",
            max_events=self.parquet_batching_max_events,
            flush_after_seconds=self.parquet_batching_timeout_secs,
            columns=[
                StatsData.TIMESTAMP,
                StatsData.STATS,
                WriterEvent.ENDPOINT_ID,
                StatsData.STATS_NAME,
            ],
            partition_cols=[WriterEvent.ENDPOINT_ID, StatsData.STATS_NAME],
            single_file=True,
        )


class ReconstructWriterEvent(storey.MapClass):
    def __init__(self):
        super().__init__()

    def do(self, event: dict) -> dict[str, Any]:
        logger.info("Reconstructing the event", event=event)
        kind = event.pop(WriterEvent.EVENT_KIND, WriterEventKind.RESULT)
        result_event = _AppResultEvent(json.loads(event.pop(WriterEvent.DATA, "{}")))
        result_event.update(_AppResultEvent(event))

        expected_keys = list(
            set(WriterEvent.list()).difference(
                [WriterEvent.EVENT_KIND, WriterEvent.DATA]
            )
        )
        if kind == WriterEventKind.METRIC:
            expected_keys.extend(MetricData.list())
        elif kind == WriterEventKind.RESULT:
            expected_keys.extend(ResultData.list())
        elif kind == WriterEventKind.STATS:
            expected_keys.extend(StatsData.list())
        else:
            raise _WriterEventValueError(
                f"Unknown event kind: {kind}, expected one of: {WriterEventKind.list()}"
            )
        missing_keys = [key for key in expected_keys if key not in result_event]
        if missing_keys:
            raise _WriterEventValueError(
                f"The received event misses some keys compared to the expected "
                f"monitoring application event schema: {missing_keys} for event kind {kind}"
            )
        result_event["kind"] = kind
        if kind in WriterEventKind.user_app_outputs():
            result_event[WriterEvent.END_INFER_TIME] = datetime.fromisoformat(
                event[WriterEvent.END_INFER_TIME]
            )
        if kind == WriterEventKind.STATS:
            result_event[StatsData.STATS] = json.dumps(result_event[StatsData.STATS])
        return result_event


class KindChoice(storey.Choice):
    def select_outlets(self, event):
        kind = event.get("kind")
        logger.info("Selecting the outlet for the event", kind=kind)
        if kind == WriterEventKind.METRIC:
            outlets = ["tsdb_metrics"]
        elif kind == WriterEventKind.RESULT:
            outlets = ["tsdb_app_results", "alert_generator"]
        elif kind == WriterEventKind.STATS:
            outlets = ["stats_writer"]
        else:
            raise _WriterEventValueError(
                f"Unknown event kind: {kind}, expected one of: {WriterEventKind.list()}"
            )
        return outlets


class AlertGenerator(storey.MapClass):
    def __init__(self, project: str, **kwargs):
        self.project = project
        super().__init__(**kwargs)

    def do(self, event: dict) -> Optional[dict[str, Any]]:
        kind = event.pop(WriterEvent.EVENT_KIND, WriterEventKind.RESULT)
        if (
            mlrun.mlconf.alerts.mode == mlrun.common.schemas.alert.AlertsModes.enabled
            and kind == WriterEventKind.RESULT
            and (
                event[ResultData.RESULT_STATUS] == ResultStatusApp.detected.value
                or event[ResultData.RESULT_STATUS]
                == ResultStatusApp.potential_detection.value
            )
        ):
            event_value = {
                "app_name": event[WriterEvent.APPLICATION_NAME],
                "model": event[WriterEvent.ENDPOINT_NAME],
                "model_endpoint_id": event[WriterEvent.ENDPOINT_ID],
                "result_name": event[ResultData.RESULT_NAME],
                "result_value": event[ResultData.RESULT_VALUE],
            }
            data = self._generate_event_data(
                entity_id=get_result_instance_fqn(
                    event[WriterEvent.ENDPOINT_ID],
                    event[WriterEvent.APPLICATION_NAME],
                    event[ResultData.RESULT_NAME],
                ),
                result_status=event[ResultData.RESULT_STATUS],
                event_value=event_value,
                project_name=self.project,
                result_kind=event[ResultData.RESULT_KIND],
            )
            event = data.dict()
            logger.info("Generated alert event", event=event)
            return event
        return None

    @staticmethod
    def _generate_alert_event_kind(
        result_kind: int, result_status: int
    ) -> alert_objects.EventKind:
        """Generate the required Event Kind format for the alerting system"""
        event_kind = ResultKindApp(value=result_kind).name

        if result_status == ResultStatusApp.detected.value:
            event_kind = f"{event_kind}_detected"
        else:
            event_kind = f"{event_kind}_suspected"
        return alert_objects.EventKind(
            value=mlrun.utils.helpers.normalize_name(event_kind)
        )

    def _generate_event_data(
        self,
        entity_id: str,
        result_status: int,
        event_value: dict,
        project_name: str,
        result_kind: int,
    ) -> mlrun.common.schemas.Event:
        entity = mlrun.common.schemas.alert.EventEntities(
            kind=alert_objects.EventEntityKind.MODEL_ENDPOINT_RESULT,
            project=project_name,
            ids=[entity_id],
        )

        event_kind = self._generate_alert_event_kind(
            result_status=result_status, result_kind=result_kind
        )

        event_data = mlrun.common.schemas.Event(
            kind=alert_objects.EventKind(value=event_kind),
            entity=entity,
            value_dict=event_value,
        )

        return event_data



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Don't remove this, used by sphinx documentation
__all__ = [
    "load_project",
    "new_project",
    "get_or_create_project",
    "MlrunProject",
    "ProjectMetadata",
    "ProjectSpec",
    "ProjectStatus",
    "run_function",
    "build_function",
    "deploy_function",
]

from .operations import build_function, deploy_function, run_function  # noqa
from .pipelines import (
    import_remote_project,
    load_and_run_workflow,
    load_and_run,
    pipeline_context,
    rerun_workflow,
)  # noqa
from .project import (
    MlrunProject,
    ProjectMetadata,
    ProjectSpec,
    ProjectStatus,
    get_or_create_project,
    load_project,
    new_project,
)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import socket
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterator
from contextlib import contextmanager, nullcontext
from datetime import datetime, timedelta, timezone
from typing import Any, Literal, Optional, Union, cast

import pandas as pd

import mlrun
import mlrun.common.constants as mlrun_constants
import mlrun.common.helpers
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.common.types
import mlrun.datastore.datastore_profile as ds_profile
import mlrun.errors
import mlrun.model_monitoring.api as mm_api
import mlrun.model_monitoring.applications.context as mm_context
import mlrun.model_monitoring.applications.results as mm_results
import mlrun.model_monitoring.db._schedules as mm_schedules
import mlrun.model_monitoring.helpers as mm_helpers
import mlrun.utils
from mlrun.serving.utils import MonitoringApplicationToDict
from mlrun.utils import logger


class ExistingDataHandling(mlrun.common.types.StrEnum):
    fail_on_overlap = "fail_on_overlap"
    skip_overlap = "skip_overlap"
    delete_all = "delete_all"


def _serialize_context_and_result(
    *,
    context: mm_context.MonitoringApplicationContext,
    result: Union[
        mm_results.ModelMonitoringApplicationResult,
        mm_results.ModelMonitoringApplicationMetric,
        mm_results._ModelMonitoringApplicationStats,
    ],
) -> dict[mm_constants.WriterEvent, str]:
    """
    Serialize the returned result from a model monitoring application and its context
    for the writer.
    """
    writer_event = {
        mm_constants.WriterEvent.ENDPOINT_NAME: context.endpoint_name,
        mm_constants.WriterEvent.APPLICATION_NAME: context.application_name,
        mm_constants.WriterEvent.ENDPOINT_ID: context.endpoint_id,
        mm_constants.WriterEvent.START_INFER_TIME: context.start_infer_time.isoformat(
            sep=" ", timespec="microseconds"
        ),
        mm_constants.WriterEvent.END_INFER_TIME: context.end_infer_time.isoformat(
            sep=" ", timespec="microseconds"
        ),
    }

    if isinstance(result, mm_results.ModelMonitoringApplicationResult):
        writer_event[mm_constants.WriterEvent.EVENT_KIND] = (
            mm_constants.WriterEventKind.RESULT
        )
    elif isinstance(result, mm_results._ModelMonitoringApplicationStats):
        writer_event[mm_constants.WriterEvent.EVENT_KIND] = (
            mm_constants.WriterEventKind.STATS
        )
    else:
        writer_event[mm_constants.WriterEvent.EVENT_KIND] = (
            mm_constants.WriterEventKind.METRIC
        )
    writer_event[mm_constants.WriterEvent.DATA] = json.dumps(result.to_dict())

    return writer_event


class ModelMonitoringApplicationBase(MonitoringApplicationToDict, ABC):
    """
    The base class for a model monitoring application.
    Inherit from this class to create a custom model monitoring application.

    For example, :code:`MyApp` below is a simplistic custom application::

        from mlrun.common.schemas.model_monitoring.constants import (
            ResultKindApp,
            ResultStatusApp,
        )
        from mlrun.model_monitoring.applications import (
            ModelMonitoringApplicationBase,
            ModelMonitoringApplicationResult,
            MonitoringApplicationContext,
        )


        class MyApp(ModelMonitoringApplicationBase):
            def do_tracking(
                self, monitoring_context: MonitoringApplicationContext
            ) -> ModelMonitoringApplicationResult:
                monitoring_context.logger.info(
                    "Running application",
                    application_name=monitoring_context.application_name,
                )
                return ModelMonitoringApplicationResult(
                    name="data_drift_test",
                    value=0.5,
                    kind=ResultKindApp.data_drift,
                    status=ResultStatusApp.detected,
                )
    """

    kind = "monitoring_application"

    def do(
        self, monitoring_context: mm_context.MonitoringApplicationContext
    ) -> tuple[
        list[
            Union[
                mm_results.ModelMonitoringApplicationResult,
                mm_results.ModelMonitoringApplicationMetric,
            ]
        ],
        mm_context.MonitoringApplicationContext,
    ]:
        """
        Process the monitoring event and return application results & metrics.
        Note: this method is internal and should not be called directly or overridden.

        :param monitoring_context:   (MonitoringApplicationContext) The monitoring application context.
        :returns:                    A tuple of:
                                        [0] = list of application results that can be either from type
                                        `ModelMonitoringApplicationResult`
                                        or from type `ModelMonitoringApplicationResult`.
                                        [1] = the original application event, wrapped in `MonitoringApplicationContext`
                                         object
        """
        results = self.do_tracking(monitoring_context=monitoring_context)
        if isinstance(results, dict):
            results = [
                mm_results.ModelMonitoringApplicationMetric(name=key, value=value)
                for key, value in results.items()
            ]
        results = results if isinstance(results, list) else [results]
        return results, monitoring_context

    @staticmethod
    def _flatten_data_result(
        result: Union[
            list[mm_results._ModelMonitoringApplicationDataRes],
            mm_results._ModelMonitoringApplicationDataRes,
        ],
    ) -> Union[list[dict], dict]:
        """Flatten result/metric objects to dictionaries"""
        if isinstance(result, mm_results._ModelMonitoringApplicationDataRes):
            return result.to_dict()
        if isinstance(result, list):
            return [
                element.to_dict()
                if isinstance(element, mm_results._ModelMonitoringApplicationDataRes)
                else element
                for element in result
            ]
        return result

    @staticmethod
    def _check_writer_is_up(project: "mlrun.MlrunProject") -> None:
        try:
            project.get_function(
                mm_constants.MonitoringFunctionNames.WRITER, ignore_cache=True
            )
        except mlrun.errors.MLRunNotFoundError:
            raise mlrun.errors.MLRunValueError(
                "Writing outputs to the databases is blocked as the model monitoring infrastructure is disabled.\n"
                "To unblock, enable model monitoring with `project.enable_model_monitoring()`."
            )

    @classmethod
    @contextmanager
    def _push_to_writer(
        cls,
        *,
        write_output: bool,
        application_name: str,
        artifact_path: str,
        stream_profile: Optional[ds_profile.DatastoreProfile],
        project: "mlrun.MlrunProject",
    ) -> Iterator[
        tuple[
            dict[str, list[tuple]],
            Optional[mm_schedules.ModelMonitoringSchedulesFileApplication],
        ]
    ]:
        endpoints_output: dict[
            str,
            list[
                tuple[
                    mm_context.MonitoringApplicationContext,
                    Union[
                        mm_results.ModelMonitoringApplicationResult,
                        mm_results.ModelMonitoringApplicationMetric,
                        list[
                            Union[
                                mm_results.ModelMonitoringApplicationResult,
                                mm_results.ModelMonitoringApplicationMetric,
                                mm_results._ModelMonitoringApplicationStats,
                            ]
                        ],
                    ],
                ]
            ],
        ] = defaultdict(list)
        application_schedules = nullcontext()
        if write_output:
            cls._check_writer_is_up(project)
            application_schedules = (
                mm_schedules.ModelMonitoringSchedulesFileApplication(
                    artifact_path, application=application_name
                )
            )
        try:
            yield endpoints_output, application_schedules.__enter__()
        finally:
            if write_output and any(endpoints_output.values()):
                logger.debug(
                    "Pushing model monitoring application job data to the writer stream",
                    passed_stream_profile=str(stream_profile),
                )
                project_name = (
                    mlrun.mlconf.active_project or mlrun.get_current_project().name
                )
                writer_stream = mm_helpers.get_output_stream(
                    project=project_name,
                    function_name=mm_constants.MonitoringFunctionNames.WRITER,
                    profile=stream_profile,
                )
                for endpoint_id, outputs in endpoints_output.items():
                    writer_events = []
                    for ctx, res in outputs:
                        if isinstance(res, list):
                            writer_events.extend(
                                _serialize_context_and_result(
                                    context=ctx, result=sub_res
                                )
                                for sub_res in res
                            )
                        else:
                            writer_events.append(
                                _serialize_context_and_result(context=ctx, result=res)
                            )
                    writer_stream.push(
                        writer_events,
                        partition_key=endpoint_id,
                    )
                logger.debug(
                    "Pushed the data to all the relevant model endpoints successfully",
                    endpoints_output=endpoints_output,
                )

                logger.debug(
                    "Saving the application schedules",
                    application_name=application_name,
                )
                application_schedules.__exit__(None, None, None)

    @classmethod
    def _get_application_name(cls, context: "mlrun.MLClientCtx") -> str:
        """Get the application name from the context via the function URI"""
        _, application_name, _, _ = mlrun.common.helpers.parse_versioned_object_uri(
            context.to_dict().get("spec", {}).get("function", "")
        )
        return application_name

    def _handler(
        self,
        context: "mlrun.MLClientCtx",
        sample_data: Optional[pd.DataFrame] = None,
        reference_data: Optional[pd.DataFrame] = None,
        endpoints: Union[
            list[tuple[str, str]], list[list[str]], list[str], Literal["all"], None
        ] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        base_period: Optional[int] = None,
        write_output: bool = False,
        existing_data_handling: ExistingDataHandling = ExistingDataHandling.fail_on_overlap,
        stream_profile: Optional[ds_profile.DatastoreProfile] = None,
    ):
        """
        A custom handler that wraps the application's logic implemented in
        :py:meth:`~mlrun.model_monitoring.applications.ModelMonitoringApplicationBase.do_tracking`
        for an MLRun job.
        This method should not be called directly.
        """
        project = context.get_project_object()
        if not project:
            raise mlrun.errors.MLRunValueError("Could not load project from context")

        if write_output and (
            not endpoints or sample_data is not None or reference_data is not None
        ):
            raise mlrun.errors.MLRunValueError(
                "Writing the results of an application to the TSDB is possible only when "
                "working with endpoints, without any custom data-frame input"
            )

        application_name = self._get_application_name(context)

        feature_stats = (
            mm_api.get_sample_set_statistics(reference_data)
            if reference_data is not None
            else None
        )

        with self._push_to_writer(
            write_output=write_output,
            stream_profile=stream_profile,
            application_name=application_name,
            artifact_path=context.artifact_path,
            project=project,
        ) as (endpoints_output, application_schedules):

            def call_do_tracking(
                monitoring_context: mm_context.MonitoringApplicationContext,
            ):
                nonlocal endpoints_output

                result = self.do_tracking(monitoring_context)
                endpoints_output[monitoring_context.endpoint_id].append(
                    (monitoring_context, result)
                )
                return result

            if endpoints is not None:
                resolved_endpoints = self._normalize_and_validate_endpoints(
                    project=project, endpoints=endpoints
                )
                if (
                    write_output
                    and existing_data_handling == ExistingDataHandling.delete_all
                ):
                    endpoint_ids = [
                        endpoint_id for _, endpoint_id in resolved_endpoints
                    ]
                    context.logger.info(
                        "Deleting all the application data before running the application",
                        application_name=application_name,
                        endpoint_ids=endpoint_ids,
                    )
                    self._delete_application_data(
                        project_name=project.name,
                        application_name=application_name,
                        endpoint_ids=endpoint_ids,
                        application_schedules=application_schedules,
                    )
                for endpoint_name, endpoint_id in resolved_endpoints:
                    for monitoring_ctx in self._window_generator(
                        start=start,
                        end=end,
                        base_period=base_period,
                        application_schedules=application_schedules,
                        endpoint_id=endpoint_id,
                        endpoint_name=endpoint_name,
                        application_name=application_name,
                        existing_data_handling=existing_data_handling,
                        sample_data=sample_data,
                        context=context,
                        project=project,
                    ):
                        result = call_do_tracking(monitoring_ctx)
                        result_key = (
                            f"{endpoint_name}-{endpoint_id}_{monitoring_ctx.start_infer_time.isoformat()}_{monitoring_ctx.end_infer_time.isoformat()}"
                            if monitoring_ctx.start_infer_time
                            and monitoring_ctx.end_infer_time
                            else f"{endpoint_name}-{endpoint_id}"
                        )

                        context.log_result(
                            result_key, self._flatten_data_result(result)
                        )
                # Check if no result was produced for any endpoint (e.g., due to no data in all windows)
                if not any(endpoints_output.values()):
                    context.logger.warning(
                        "No data was found for any of the specified endpoints. "
                        "No results were produced",
                        application_name=application_name,
                        endpoints=endpoints,
                        start=start,
                        end=end,
                    )
            else:
                result = call_do_tracking(
                    mm_context.MonitoringApplicationContext._from_ml_ctx(
                        context=context,
                        project=project,
                        application_name=application_name,
                        event={},
                        sample_df=sample_data,
                        feature_stats=feature_stats,
                    )
                )
                return self._flatten_data_result(result)

    @staticmethod
    def _check_endpoints_first_request(
        endpoints: list[mlrun.common.schemas.ModelEndpoint],
    ) -> None:
        """Make sure that all the endpoints have had at least one request"""
        endpoints_no_requests = [
            (endpoint.metadata.name, endpoint.metadata.uid)
            for endpoint in endpoints
            if not endpoint.status.first_request
        ]
        if endpoints_no_requests:
            raise mlrun.errors.MLRunValueError(
                "The following model endpoints have not had any requests yet and "
                "have no data, cannot run the model monitoring application on them: "
                f"{endpoints_no_requests}"
            )

    @classmethod
    def _normalize_and_validate_endpoints(
        cls,
        project: "mlrun.MlrunProject",
        endpoints: Union[
            list[tuple[str, str]], list[list[str]], list[str], Literal["all"]
        ],
    ) -> list[tuple[str, str]]:
        if isinstance(endpoints, list):
            if all(
                isinstance(endpoint, (tuple, list)) and len(endpoint) == 2
                for endpoint in endpoints
            ):
                # A list of [(name, uid), ...] / [[name, uid], ...] tuples/lists
                endpoint_uids_to_names = {
                    endpoint[1]: endpoint[0] for endpoint in endpoints
                }
                endpoints_list = project.list_model_endpoints(
                    uids=list(endpoint_uids_to_names.keys()), latest_only=True
                ).endpoints

                # Check for missing endpoint uids or name/uid mismatches
                for endpoint in endpoints_list:
                    if (
                        endpoint_uids_to_names[cast(str, endpoint.metadata.uid)]
                        != endpoint.metadata.name
                    ):
                        raise mlrun.errors.MLRunNotFoundError(
                            "Could not find model endpoint with name "
                            f"'{endpoint_uids_to_names[cast(str, endpoint.metadata.uid)]}' "
                            f"and uid '{endpoint.metadata.uid}'"
                        )
                missing = set(endpoint_uids_to_names.keys()) - {
                    cast(str, endpoint.metadata.uid) for endpoint in endpoints_list
                }
                if missing:
                    raise mlrun.errors.MLRunNotFoundError(
                        "Could not find model endpoints with the following uids: "
                        f"{missing}"
                    )

            elif all(isinstance(endpoint, str) for endpoint in endpoints):
                # A list of [name, ...] strings
                endpoint_names = cast(list[str], endpoints)
                endpoints_list = project.list_model_endpoints(
                    names=endpoint_names, latest_only=True
                ).endpoints

                # Check for missing endpoint names
                missing = set(endpoints) - {
                    endpoint.metadata.name for endpoint in endpoints_list
                }
                if missing:
                    logger.warning(
                        "Could not list all the required endpoints",
                        missing_endpoints=missing,
                        endpoints_list=endpoints_list,
                    )
            else:
                raise mlrun.errors.MLRunValueError(
                    "Could not resolve the following list as a list of endpoints:\n"
                    f"{endpoints}\n"
                    "The list must be either a list of (name, uid) tuples/lists or a list of names."
                )
        elif endpoints == "all":
            endpoints_list = project.list_model_endpoints(latest_only=True).endpoints
        elif isinstance(endpoints, str):
            raise mlrun.errors.MLRunValueError(
                'A string input for `endpoints` can only be "all" for all the model endpoints in '
                "the project. If you want to select a single model endpoint with the given name, "
                f'use a list: `endpoints=["{endpoints}"]`.'
            )
        else:
            raise mlrun.errors.MLRunValueError(
                "Could not resolve the `endpoints` parameter. The parameter must be either:\n"
                "- a list of (name, uid) tuples/lists\n"
                "- a list of names\n"
                '- the string "all" for all the model endpoints in the project.'
            )

        if not endpoints_list:
            raise mlrun.errors.MLRunNotFoundError(
                f"Did not find any model endpoints {endpoints=}"
            )

        cls._check_endpoints_first_request(endpoints_list)

        return [
            (endpoint.metadata.name, cast(str, endpoint.metadata.uid))
            for endpoint in endpoints_list
        ]

    @staticmethod
    def _validate_and_get_window_length(
        *, base_period: int, start_dt: datetime, end_dt: datetime
    ) -> timedelta:
        if not isinstance(base_period, int) or base_period <= 0:
            raise mlrun.errors.MLRunValueError(
                "`base_period` must be a nonnegative integer - the number of minutes in a monitoring window"
            )

        window_length = timedelta(minutes=base_period)

        full_interval_length = end_dt - start_dt
        remainder = full_interval_length % window_length
        if remainder:
            if full_interval_length < window_length:
                extra_msg = (
                    "The `base_period` is longer than the difference between `end` and `start`: "
                    f"{full_interval_length}. Consider not specifying `base_period`."
                )
            else:
                extra_msg = (
                    f"Consider changing the `end` time to `end`={end_dt - remainder}"
                )
            raise mlrun.errors.MLRunValueError(
                "The difference between `end` and `start` must be a multiple of `base_period`: "
                f"`base_period`={window_length}, `start`={start_dt}, `end`={end_dt}. "
                f"{extra_msg}"
            )
        return window_length

    @staticmethod
    def _validate_monotonically_increasing_data(
        *,
        application_schedules: Optional[
            mm_schedules.ModelMonitoringSchedulesFileApplication
        ],
        endpoint_id: str,
        start_dt: datetime,
        end_dt: datetime,
        base_period: Optional[int],
        application_name: str,
        existing_data_handling: ExistingDataHandling,
    ) -> datetime:
        """Make sure that the (app, endpoint) pair doesn't write output before the last analyzed window"""
        if application_schedules:
            last_analyzed = application_schedules.get_endpoint_last_analyzed(
                endpoint_id
            )
            if last_analyzed:
                if start_dt < last_analyzed:
                    if existing_data_handling == ExistingDataHandling.skip_overlap:
                        if last_analyzed < end_dt and base_period is None:
                            logger.warn(
                                "Setting the start time to last_analyzed since the original start time precedes "
                                "last_analyzed",
                                original_start=start_dt,
                                new_start=last_analyzed,
                                application_name=application_name,
                                endpoint_id=endpoint_id,
                            )
                            start_dt = last_analyzed
                        else:
                            raise mlrun.errors.MLRunValueError(
                                "The start time for the application and endpoint precedes the last analyzed time: "
                                f"start_dt='{start_dt}', last_analyzed='{last_analyzed}', {application_name=}, "
                                f"{endpoint_id=}. "
                                "Writing data out of order is not supported, and the start time could not be "
                                "dynamically reset, as last_analyzed is later than the given end time or that "
                                f"base_period was specified (end_dt='{end_dt}', {base_period=})."
                            )
                    else:
                        raise mlrun.errors.MLRunValueError(
                            "The start time for the application and endpoint precedes the last analyzed time: "
                            f"start_dt='{start_dt}', last_analyzed='{last_analyzed}', {application_name=}, "
                            f"{endpoint_id=}. "
                            "Writing data out of order is not supported. You should change the start time to "
                            f"'{last_analyzed}' or later."
                        )
            else:
                logger.debug(
                    "The application is running on the endpoint for the first time",
                    endpoint_id=endpoint_id,
                    start_dt=start_dt,
                    application_name=application_name,
                )
        return start_dt

    @staticmethod
    def _delete_application_data(
        project_name: str,
        application_name: str,
        endpoint_ids: list[str],
        application_schedules: Optional[
            mm_schedules.ModelMonitoringSchedulesFileApplication
        ],
    ) -> None:
        mlrun.get_run_db().delete_model_monitoring_metrics(
            project=project_name,
            application_name=application_name,
            endpoint_ids=endpoint_ids,
        )
        if application_schedules:
            application_schedules.delete_endpoints_last_analyzed(
                endpoint_uids=endpoint_ids
            )

    @classmethod
    def _window_generator(
        cls,
        *,
        start: Optional[str],
        end: Optional[str],
        base_period: Optional[int],
        application_schedules: Optional[
            mm_schedules.ModelMonitoringSchedulesFileApplication
        ],
        endpoint_name: str,
        endpoint_id: str,
        application_name: str,
        existing_data_handling: ExistingDataHandling,
        context: "mlrun.MLClientCtx",
        project: "mlrun.MlrunProject",
        sample_data: Optional[pd.DataFrame],
    ) -> Iterator[mm_context.MonitoringApplicationContext]:
        def yield_monitoring_ctx(
            window_start: Optional[datetime], window_end: Optional[datetime]
        ) -> Iterator[mm_context.MonitoringApplicationContext]:
            ctx = mm_context.MonitoringApplicationContext._from_ml_ctx(
                event={
                    mm_constants.ApplicationEvent.ENDPOINT_NAME: endpoint_name,
                    mm_constants.ApplicationEvent.ENDPOINT_ID: endpoint_id,
                    mm_constants.ApplicationEvent.START_INFER_TIME: window_start,
                    mm_constants.ApplicationEvent.END_INFER_TIME: window_end,
                },
                application_name=application_name,
                context=context,
                project=project,
                sample_df=sample_data,
            )

            if ctx.sample_df.empty:
                # The current sample is empty
                context.logger.debug(
                    "No sample data available for tracking",
                    application_name=application_name,
                    endpoint_id=ctx.endpoint_id,
                    start_time=ctx.start_infer_time,
                    end_time=ctx.end_infer_time,
                )
                return

            yield ctx

            if application_schedules and window_end:
                application_schedules.update_endpoint_last_analyzed(
                    endpoint_uid=endpoint_id, last_analyzed=window_end
                )

        if start is None or end is None:
            # A single window based on the `sample_data` input - see `_handler`.
            yield from yield_monitoring_ctx(None, None)
            return

        start_dt = datetime.fromisoformat(start)
        end_dt = datetime.fromisoformat(end)

        # If `start_dt` and `end_dt` do not include time zone information - change them to UTC
        if (start_dt.tzinfo is None) and (end_dt.tzinfo is None):
            start_dt = start_dt.replace(tzinfo=timezone.utc)
            end_dt = end_dt.replace(tzinfo=timezone.utc)
        elif (start_dt.tzinfo is None) or (end_dt.tzinfo is None):
            raise mlrun.errors.MLRunValueError(
                "The start and end times must either both include time zone information or both be naive (no time "
                f"zone). Asserting the above failed, aborting the evaluate request: start={start}, end={end}."
            )

        if existing_data_handling != ExistingDataHandling.delete_all:
            start_dt = cls._validate_monotonically_increasing_data(
                application_schedules=application_schedules,
                endpoint_id=endpoint_id,
                start_dt=start_dt,
                end_dt=end_dt,
                base_period=base_period,
                application_name=application_name,
                existing_data_handling=existing_data_handling,
            )

        if base_period is None:
            yield from yield_monitoring_ctx(start_dt, end_dt)
            return

        window_length = cls._validate_and_get_window_length(
            base_period=base_period, start_dt=start_dt, end_dt=end_dt
        )

        current_start_time = start_dt
        while current_start_time < end_dt:
            current_end_time = min(current_start_time + window_length, end_dt)
            yield from yield_monitoring_ctx(current_start_time, current_end_time)
            current_start_time = current_end_time

    @classmethod
    def deploy(
        cls,
        func_name: str,
        func_path: Optional[str] = None,
        image: Optional[str] = None,
        handler: Optional[str] = None,
        with_repo: Optional[bool] = False,
        tag: Optional[str] = None,
        requirements: Optional[Union[str, list[str]]] = None,
        requirements_file: str = "",
        **application_kwargs,
    ) -> None:
        """
        Set the application to the current project and deploy it as a Nuclio serving function.
        Required for your model monitoring application to work as a part of the model monitoring framework.

        :param func_name: The name of the function.
        :param func_path: The path of the function, :code:`None` refers to the current Jupyter notebook.

        For the other arguments, refer to
        :py:meth:`~mlrun.projects.MlrunProject.set_model_monitoring_function`.
        """
        project = cast("mlrun.MlrunProject", mlrun.get_current_project())
        function = project.set_model_monitoring_function(
            name=func_name,
            func=func_path,
            application_class=cls.__name__,
            handler=handler,
            image=image,
            with_repo=with_repo,
            requirements=requirements,
            requirements_file=requirements_file,
            tag=tag,
            **application_kwargs,
        )
        function.deploy()

    @classmethod
    def get_job_handler(cls, handler_to_class: str) -> str:
        """
        A helper function to get the handler to the application job ``_handler``.

        :param handler_to_class: The handler to the application class, e.g. ``my_package.sub_module1.MonitoringApp1``.
        :returns:                The handler to the job of the application class.
        """
        return f"{handler_to_class}::{cls._handler.__name__}"

    @classmethod
    def _determine_job_name(
        cls,
        *,
        func_name: Optional[str],
        class_handler: Optional[str],
        handler_to_class: str,
    ) -> str:
        """
        Determine the batch app's job name. This name is used also as the application name,
        which is retrieved in `_get_application_name`.
        """
        if func_name:
            job_name = func_name
        else:
            if not class_handler:
                class_name = cls.__name__
            else:
                class_name = handler_to_class.split(".")[-1].split("::")[0]

            job_name = mlrun.utils.normalize_name(class_name)

        if not mm_constants.APP_NAME_REGEX.fullmatch(job_name):
            raise mlrun.errors.MLRunValueError(
                "The function name does not comply with the required pattern "
                f"`{mm_constants.APP_NAME_REGEX.pattern}`. "
                "Please choose another `func_name`."
            )
        job_name, was_renamed, suffix = mlrun.utils.helpers.ensure_batch_job_suffix(
            job_name
        )
        if was_renamed:
            mlrun.utils.logger.info(
                f'Changing function name - adding `"{suffix}"` suffix',
                func_name=job_name,
            )

        return job_name

    @classmethod
    def to_job(
        cls,
        *,
        class_handler: Optional[str] = None,
        func_path: Optional[str] = None,
        func_name: Optional[str] = None,
        tag: Optional[str] = None,
        image: Optional[str] = None,
        with_repo: Optional[bool] = False,
        requirements: Optional[Union[str, list[str]]] = None,
        requirements_file: str = "",
        project: Optional["mlrun.MlrunProject"] = None,
    ) -> mlrun.runtimes.KubejobRuntime:
        """
        Get the application's :py:meth:`~mlrun.model_monitoring.applications.ModelMonitoringApplicationBase.do_tracking`
        model monitoring logic as a :py:class:`~mlrun.runtimes.KubejobRuntime`.

        The returned job can be run as any MLRun job with the relevant inputs and params to your application:

        .. code-block:: python

            job = ModelMonitoringApplicationBase.to_job(
                class_handler="package.module.AppClass"
            )
            job.run(inputs={}, params={}, local=False)  # Add the relevant inputs and params

        Optional inputs:

        * ``sample_data``, ``pd.DataFrame``
        * ``reference_data``, ``pd.DataFrame``

        Optional params:

        * ``endpoints``, ``list[tuple[str, str]]``
        * ``start``, ``datetime``
        * ``end``, ``datetime``
        * ``base_period``, ``int``
        * ``write_output``, ``bool``
        * ``existing_data_handling``, ``str``

        For Git sources, add the source archive to the returned job and change the handler:

        .. code-block:: python

            handler = ModelMonitoringApplicationBase.get_job_handler("module.AppClass")
            job.with_source_archive(
                "git://github.com/owner/repo.git#branch-category/specific-task",
                workdir="path/to/application/folder",
                handler=handler,
            )

        :param class_handler:     The handler to the class, e.g. ``path.to.module::MonitoringApplication``,
                                  useful when using Git sources or code from images.
                                  If ``None``, the current class, deriving from
                                  :py:class:`~mlrun.model_monitoring.applications.ModelMonitoringApplicationBase`,
                                  is used.
        :param func_path:         The path to the function. If ``None``, the current notebook is used.
        :param func_name:         The name of the function. If ``None``, the normalized class name is used
                                  (:py:meth:`mlrun.utils.helpers.normalize_name`).
                                  A ``"-batch"`` suffix is guaranteed to be added if not already there.
                                  The function name is also used as the application name to use for the results.
        :param tag:               Tag for the function.
        :param image:             Docker image to run the job on (when running remotely).
        :param with_repo:         Whether to clone the current repo to the build source.
        :param requirements:      List of Python requirements to be installed in the image.
        :param requirements_file: Path to a Python requirements file to be installed in the image.
        :param project:           The current project to set the function to. If not set, the current project is used.

        :returns: The :py:class:`~mlrun.runtimes.KubejobRuntime` job that wraps the model monitoring application's
                  logic.
        """
        project = project or cast("mlrun.MlrunProject", mlrun.get_current_project())

        if not class_handler and cls == ModelMonitoringApplicationBase:
            raise ValueError(
                "You must provide a handler to the model monitoring application class"
            )

        handler_to_class = class_handler or cls.__name__
        handler = cls.get_job_handler(handler_to_class)

        job_name = cls._determine_job_name(
            func_name=func_name,
            class_handler=class_handler,
            handler_to_class=handler_to_class,
        )

        job = cast(
            mlrun.runtimes.KubejobRuntime,
            project.set_function(
                func=func_path,
                name=job_name,
                kind=mlrun.runtimes.KubejobRuntime.kind,
                handler=handler,
                tag=tag,
                image=image,
                with_repo=with_repo,
                requirements=requirements,
                requirements_file=requirements_file,
            ),
        )
        return job

    @classmethod
    def evaluate(
        cls,
        func_path: Optional[str] = None,
        func_name: Optional[str] = None,
        *,
        tag: Optional[str] = None,
        run_local: bool = True,
        auto_build: bool = True,
        sample_data: Optional[Union[pd.DataFrame, str]] = None,
        reference_data: Optional[Union[pd.DataFrame, str]] = None,
        image: Optional[str] = None,
        with_repo: Optional[bool] = False,
        class_handler: Optional[str] = None,
        requirements: Optional[Union[str, list[str]]] = None,
        requirements_file: str = "",
        endpoints: Union[list[tuple[str, str]], list[str], Literal["all"], None] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
        base_period: Optional[int] = None,
        write_output: bool = False,
        existing_data_handling: ExistingDataHandling = ExistingDataHandling.fail_on_overlap,
        stream_profile: Optional[ds_profile.DatastoreProfile] = None,
    ) -> "mlrun.RunObject":
        """
        Call this function to run the application's
        :py:meth:`~mlrun.model_monitoring.applications.ModelMonitoringApplicationBase.do_tracking`
        model monitoring logic as a :py:class:`~mlrun.runtimes.KubejobRuntime`, which is an MLRun function.

        This function has default values for all of its arguments. You should change them when you want to pass
        data to the application.

        :param func_path:         The path to the function. If ``None``, the current notebook is used.
        :param func_name:         The name of the function. If ``None``, the normalized class name is used
                                  (:py:meth:`mlrun.utils.helpers.normalize_name`).
                                  A ``"-batch"`` suffix is guaranteed to be added if not already there.
                                  The function name is also used as the application name to use for the results.
        :param tag:               Tag for the function.
        :param run_local:         Whether to run the function locally or remotely.
        :param auto_build:        Whether to auto build the function.
        :param sample_data:       Pandas data-frame or :py:class:`~mlrun.artifacts.dataset.DatasetArtifact` URI as
                                  the current dataset.
                                  When set, it replaces the data read from the model endpoint's offline source.
        :param reference_data:    Pandas data-frame or :py:class:`~mlrun.artifacts.dataset.DatasetArtifact` URI as
                                  the reference dataset.
                                  When set, its statistics override the model endpoint's feature statistics.
                                  You do not need to have a model endpoint to use this option.
        :param image:             Docker image to run the job on (when running remotely).
        :param with_repo:         Whether to clone the current repo to the build source.
        :param class_handler:     The relative path to the class, useful when using Git sources or code from images.
        :param requirements:      List of Python requirements to be installed in the image.
        :param requirements_file: Path to a Python requirements file to be installed in the image.
        :param endpoints:         The model endpoints to get the data from. The options are:

                                  - a list of tuples of the model endpoints ``[(name, uid), ...]``
                                  - a list of model endpoint names ``[name, ...]``
                                  - ``"all"`` for all the project's model endpoints

                                  Note: a model endpoint name retrieves all the active model endpoints using this
                                  name, which may be more than one per name when the same name is used across
                                  multiple serving functions.

                                  If provided, and ``sample_data`` is not ``None``, you have to provide also the
                                  ``start`` and ``end`` times of the data to analyze from the model endpoints.
        :param start:             The start time of the endpoint's data, not included.
                                  If you want the model endpoint's data at ``start`` included, you need to subtract a
                                  small ``datetime.timedelta`` from it.
                                  Make sure to include the time zone when constructing ``datetime.datetime`` objects
                                  manually. When both ``start`` and ``end`` times do not include a time zone, they will
                                  be treated as UTC.
        :param end:               The end time of the endpoint's data, included.
                                  Please note: when ``start`` and ``end`` are set, they create a left-open time interval
                                  ("window") :math:`(\\operatorname{start}, \\operatorname{end}]` that excludes the
                                  endpoint's data at ``start`` and includes the data at ``end``:
                                  :math:`\\operatorname{start} < t \\leq \\operatorname{end}`, :math:`t` is the time
                                  taken in the window's data.
        :param base_period:       The window length in minutes. If ``None``, the whole window from ``start`` to ``end``
                                  is taken. If an integer is specified, the application is run from ``start`` to ``end``
                                  in ``base_period`` length windows:
                                  :math:`(\\operatorname{start}, \\operatorname{start} + \\operatorname{base\\_period}],
                                  (\\operatorname{start} + \\operatorname{base\\_period},
                                  \\operatorname{start} + 2\\cdot\\operatorname{base\\_period}],
                                  ..., (\\operatorname{start} +
                                  (m - 1)\\cdot\\operatorname{base\\_period}, \\operatorname{end}]`,
                                  where :math:`m` is a positive integer and :math:`\\operatorname{end} =
                                  \\operatorname{start} + m\\cdot\\operatorname{base\\_period}`.
                                  Please note that the difference between ``end`` and ``start`` must be a multiple of
                                  ``base_period``.
        :param write_output:      Whether to write the results and metrics to the time-series DB. Can be ``True`` only
                                  if ``endpoints`` are passed.
                                  Note: the model monitoring infrastructure must be up for the writing to work.
        :param existing_data_handling:
                                  How to handle the existing application data for the model endpoints when writing
                                  new data whose requested ``start`` time precedes the ``end`` time of a previous run
                                  that also wrote to the database. Relevant only when ``write_output=True``.
                                  The options are:

                                  - ``"fail_on_overlap"``: Default. An error is raised.
                                  - ``"skip_overlap"``:  the overlapping data is ignored and the
                                    time window is cut so that it starts at the earliest possible time after ``start``.
                                  - ``"delete_all"``: delete all the data that was written by the application to the
                                    model endpoints, regardless of the time window, and write the new data.

        :param stream_profile:    The stream datastore profile. It should be provided only when running locally and
                                  writing the outputs to the database (i.e., when both ``run_local`` and
                                  ``write_output`` are set to ``True``).
                                  For more details on configuring the stream profile, see
                                  :py:meth:`~mlrun.projects.MlrunProject.set_model_monitoring_credentials`.

        :returns: The output of the
                  :py:meth:`~mlrun.model_monitoring.applications.ModelMonitoringApplicationBase.do_tracking`
                  method with the given parameters and inputs, wrapped in a :py:class:`~mlrun.model.RunObject`.
        """
        project = cast("mlrun.MlrunProject", mlrun.get_current_project())

        job = cls.to_job(
            func_path=func_path,
            func_name=func_name,
            class_handler=class_handler,
            tag=tag,
            image=image,
            with_repo=with_repo,
            requirements=requirements,
            requirements_file=requirements_file,
            project=project,
        )

        params: dict[str, Union[list, str, int, None, ds_profile.DatastoreProfile]] = {}
        if endpoints:
            params["endpoints"] = endpoints
            if sample_data is None:
                if start is None or end is None:
                    raise mlrun.errors.MLRunValueError(
                        "`start` and `end` times must be provided when `endpoints` "
                        "is provided without `sample_data`"
                    )
                params["start"] = (
                    start.isoformat() if isinstance(start, datetime) else start
                )
                params["end"] = end.isoformat() if isinstance(end, datetime) else end
                params["base_period"] = base_period
        elif start or end or base_period:
            raise mlrun.errors.MLRunValueError(
                "Custom `start` and `end` times or base_period are supported only with endpoints data"
            )
        elif write_output or stream_profile:
            raise mlrun.errors.MLRunValueError(
                "Writing the application output or passing `stream_profile` are supported only with endpoints data"
            )

        params["write_output"] = write_output
        params["existing_data_handling"] = existing_data_handling
        if stream_profile:
            if not run_local:
                raise mlrun.errors.MLRunValueError(
                    "Passing a `stream_profile` is relevant only when running locally"
                )
            if not write_output:
                raise mlrun.errors.MLRunValueError(
                    "Passing a `stream_profile` is relevant only when writing the outputs"
                )
        params["stream_profile"] = stream_profile

        inputs: dict[str, str] = {}
        for data, identifier in [
            (sample_data, "sample_data"),
            (reference_data, "reference_data"),
        ]:
            if isinstance(data, str):
                inputs[identifier] = data
            elif data is not None:
                key = f"{job.metadata.name}_{identifier}"
                inputs[identifier] = project.log_dataset(
                    key,
                    data,
                    labels={
                        mlrun_constants.MLRunInternalLabels.runner_pod: socket.gethostname(),
                        mlrun_constants.MLRunInternalLabels.producer_type: "model-monitoring-job",
                        mlrun_constants.MLRunInternalLabels.app_name: func_name
                        or cls.__name__,
                    },
                ).uri

        run_result = job.run(
            local=run_local, auto_build=auto_build, params=params, inputs=inputs
        )
        return run_result

    @abstractmethod
    def do_tracking(
        self,
        monitoring_context: mm_context.MonitoringApplicationContext,
    ) -> Union[
        mm_results.ModelMonitoringApplicationResult,
        list[
            Union[
                mm_results.ModelMonitoringApplicationResult,
                mm_results.ModelMonitoringApplicationMetric,
            ]
        ],
        dict[str, Any],
    ]:
        """
        Implement this method with your custom monitoring logic.

        :param monitoring_context:      (MonitoringApplicationContext) The monitoring context to process.

        :returns:                       (ModelMonitoringApplicationResult) or
                                        (list[Union[ModelMonitoringApplicationResult,
                                        ModelMonitoringApplicationMetric]])
                                        or dict that contains the application metrics only (in this case the name of
                                        each metric name is the key and the metric value is the corresponding value).
        """
        raise NotImplementedError



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional, Union

from mlrun.artifacts import Artifact
from mlrun.datastore import DataItem

from .utils import TypeHintUtils


class Packager(ABC):
    """
    The abstract base class for a packager. Packager has two main duties:

    1. **Packing** - get an object that was returned from a function and log it to MLRun. The user can specify packing
       configurations to the packager using log hints. The packed object can be an artifact or a result.
    2. **Unpacking** - get an :py:meth:`mlrun.DataItem<mlrun.datastore.base.DataItem>` (an input to a MLRun function)
       and parse it to the desired hinted type. The packager uses the instructions it noted itself when originally
       packing the object.

    .. rubric:: Custom Implementation (Inherit Packager)

    The Packager has one class variable and five class methods that must be implemented:

    * :py:meth:`PACKABLE_OBJECT_TYPE<PACKABLE_OBJECT_TYPE>` - A class variable to specify the object type this packager
      handles. Used for the ``is_packable`` and ``repr`` methods. An ellipses (`...`) means any type.
    * :py:meth:`PRIORITY<PRIORITY>` - The priority of this packager among the rest of the packagers. Valid values are
      integers between 1-10 where 1 is the highest priority and 10 is the lowest. If not set, a default priority of 5 is
      set for MLRun builtin packagers and 3 for user custom packagers.
    * :py:meth:`get_default_packing_artifact_type` - A class method to get the default artifact type for packing an
      object when it is not provided by the user.
    * :py:meth:`get_default_unpacking_artifact_type` - A class method to get the default artifact type for unpacking a
      data item when it is not representing a package, but a simple url or an old / manually logged artifact.
    * :py:meth:`get_supported_artifact_types` - A class method to get the supported artifact types this packager can
      pack an object as. Used for the ``is_packable`` and `repr` methods.
    * :py:meth:`pack` - A class method to pack a returned object using the provided log hint configurations while noting
      itself instructions for how to unpack it once needed (only relevant for packed artifacts since results do not need
      unpacking).
    * :py:meth:`unpack` - A class method to unpack an MLRun ``DataItem``, parsing it to its desired hinted type using
      the instructions noted while originally packing it.

    The class methods ``is_packable`` and ``is_unpackable`` are implemented with the following basic logic:

    * :py:meth:`is_packable` - a class method to know whether to use this packager to pack an object by its
      type and artifact type. It compares the object's type with the ``PACKABLE_OBJECT_TYPE`` and checks that the
      artifact type is in the returned supported artifacts list from ``get_supported_artifact_types``.
    * :py:meth:`is_unpackable` - a class method to know whether to use this packager to unpack a data item by the user-
      noted type hint and optionally stored artifact type in the data item (in case it was packaged before). It matches
      the ``PACKABLE_OBJECT_TYPE`` to the type hint given (same logic as IDE matchups, meaning subclasses are
      considered as unpackable) and checks if the artifact type is in the returned supported artifacts list from
      ``get_supported_artifact_types``.

    Preferably, each packager should handle a single type of object.

    .. rubric:: Linking Artifacts (extra data)

    To link between packages (using the extra data or metrics spec attributes of an artifact), use
    the key as if it exists and as value ellipses (...). The manager links all packages once it is done packing.

    For example, given extra data keys in the log hint as `extra_data`, setting them to an artifact would be::

        artifact = Artifact(key="my_artifact")
        artifact.spec.extra_data = {key: ... for key in extra_data}

    .. rubric:: Clearing Outputs

    Some of the packagers may produce files and temporary directories that should be deleted once the artifacts
    are logged. The packager can mark paths of files and directories to delete after logging using the class method
    :py:meth:`add_future_clearing_path`.

    For example, in the following packager's ``pack`` method, you can write a text file, create an Artifact, and
    then mark the text file to be deleted once the artifact is logged::

        with open("./some_file.txt", "w") as file:
            file.write("Pack me")
        artifact = Artifact(key="my_artifact")
        self.add_future_clearing_path(path="./some_file.txt")
        return artifact, None
    """

    #: The type of object this packager can pack and unpack.
    PACKABLE_OBJECT_TYPE: type = ...

    #: The priority of this packager in the packagers collection of the manager (lower is better).
    PRIORITY: int = ...

    def __init__(self):
        # Assign the packager's priority (notice that if it is equal to `...` then it will bbe overriden by the packager
        # manager when collected):
        self._priority = Packager.PRIORITY

        # List of all paths to be deleted by the manager of this packager after logging the packages:
        self._future_clearing_path_list: list[str] = []

    @abstractmethod
    def get_default_packing_artifact_type(self, obj: Any) -> str:
        """
        Get the default artifact type used for packing. The method is used when an object is sent for packing
        without an artifact type noted by the user.

        :param obj: The about to be packed object.

        :return: The default artifact type.
        """
        pass

    @abstractmethod
    def get_default_unpacking_artifact_type(self, data_item: DataItem) -> str:
        """
        Get the default artifact type used for unpacking a data item holding an object of this packager. The method
        is used when a data item is sent for unpacking without it being a package, but is a simple url or an old
        / manually logged artifact.

        :param data_item: The about-to-be unpacked data item.

        :return: The default artifact type.
        """
        pass

    @abstractmethod
    def get_supported_artifact_types(self) -> list[str]:
        """
        Get all the supported artifact types on this packager.

        :return: A list of all the supported artifact types.
        """
        pass

    @abstractmethod
    def pack(
        self,
        obj: Any,
        key: Optional[str] = None,
        artifact_type: Optional[str] = None,
        configurations: Optional[dict] = None,
    ) -> Union[tuple[Artifact, dict], dict]:
        """
        Pack an object as the given artifact type using the provided configurations.

        :param obj:            The object to pack.
        :param key:            The key of the artifact.
        :param artifact_type:  Artifact type to log to MLRun.
        :param configurations: Log hints configurations to pass to the packing method.

        :return: If the packed object is an artifact, a tuple of the packed artifact and unpacking instructions
                 dictionary. If the packed object is a result, a dictionary containing the result key and value.
        """
        pass

    @abstractmethod
    def unpack(
        self,
        data_item: DataItem,
        artifact_type: Optional[str] = None,
        instructions: Optional[dict] = None,
    ) -> Any:
        """
        Unpack the data item's artifact by the provided type using the given instructions.

        :param data_item:     The data input to unpack.
        :param artifact_type: The artifact type to unpack the data item as.
        :param instructions:  Additional instructions noted in the package to pass to the unpacking method.

        :return: The unpacked data item's object.
        """
        pass

    def is_packable(
        self,
        obj: Any,
        artifact_type: Optional[str] = None,
        configurations: Optional[dict] = None,
    ) -> bool:
        """
        Check if this packager can pack an object of the provided type as the provided artifact type.

        The default implementation checks if the packable object type of this packager is equal to the given object's
        type. If it matches, it looks for the artifact type in the list returned from
        `get_supported_artifact_types`.

        :param obj:            The object to pack.
        :param artifact_type:  The artifact type to log the object as.
        :param configurations: The log hint configurations passed by the user.

        :return: True if packable and False otherwise.
        """
        # Get the object's type:
        object_type = type(obj)

        # Validate the object type (ellipses means any type):
        if (
            self.PACKABLE_OBJECT_TYPE is not ...
            and object_type != self.PACKABLE_OBJECT_TYPE
        ):
            return False

        # Validate the artifact type (if given):
        if artifact_type and artifact_type not in self.get_supported_artifact_types():
            return False

        return True

    def is_unpackable(
        self, data_item: DataItem, type_hint: type, artifact_type: Optional[str] = None
    ) -> bool:
        """
        Check if this packager can unpack an input according to the user-given type hint and the provided artifact type.

        The default implementation tries to match the packable object type of this packager to the given type hint. If
        it matches, it looks for the artifact type in the list returned from `get_supported_artifact_types`.

        :param data_item:     The input data item to check if unpackable.
        :param type_hint:     The type hint of the input to unpack (the object type to be unpacked).
        :param artifact_type: The artifact type to unpack the object as.

        :return: True if unpackable and False otherwise.
        """
        # Check type (ellipses means any type):
        if self.PACKABLE_OBJECT_TYPE is not ...:
            if not TypeHintUtils.is_matching(
                object_type=type_hint,  # The type hint is the expected object type the MLRun function wants.
                type_hint=self.PACKABLE_OBJECT_TYPE,
                reduce_type_hint=False,
            ):
                return False

        # Check the artifact type:
        if artifact_type and artifact_type not in self.get_supported_artifact_types():
            return False

        # Unpackable:
        return True

    def add_future_clearing_path(self, path: Union[str, Path]):
        """
        Mark a path to be cleared by this packager's manager after logging the packaged artifacts.

        :param path: The path to clear post logging the artifacts.
        """
        self._future_clearing_path_list.append(str(path))

    @property
    def priority(self) -> int:
        """
        Get the packager's priority.

        :return: The packager's priority.
        """
        return self._priority

    @priority.setter
    def priority(self, priority: int):
        """
        Set the packager's priority.

        :param priority: The priority to set.
        """
        self._priority = priority

    @property
    def future_clearing_path_list(self) -> list[str]:
        """
        Get the packager's future clearing path list.

        :return: The clearing path list.
        """
        return self._future_clearing_path_list

    def __lt__(self, other: "Packager") -> bool:
        """
        A less than implementation to compare by priority in order to be able to sort the packagers by it.

        :param other: The compared packager.

        :return: True if priority is lower (means better) and False otherwise.
        """
        return self.priority < other.priority

    def __repr__(self) -> str:
        """
        Get the string representation of a packager in the following format:
        <packager name>(type=<handled type>, artifact_types=[<all supported artifact types>], priority=<priority>)

        :return: The string representation of e packager.
        """
        # Get the packager info into variables:
        packager_name = self.__class__.__name__
        handled_type = (
            (
                # Types have __name__ attribute but typing's types do not.
                self.PACKABLE_OBJECT_TYPE.__name__
                if hasattr(self.PACKABLE_OBJECT_TYPE, "__name__")
                else str(self.PACKABLE_OBJECT_TYPE)
            )
            if self.PACKABLE_OBJECT_TYPE is not ...
            else "Any"
        )
        supported_artifact_types = self.get_supported_artifact_types()

        # Return the string representation in the format noted above:
        return (
            f"{packager_name}(packable_type={handled_type}, artifact_types={supported_artifact_types}, "
            f"priority={self.priority})"
        )

    def get_data_item_local_path(
        self, data_item: DataItem, add_to_future_clearing_path: Optional[bool] = None
    ) -> str:
        """
        Get the local path to the item handled by the data item provided. The local path can be the same as the data
        item in case the data item points to a local path, or will be downloaded to a temporary directory and return
        this newly created temporary local path.

        :param data_item:                   The data item to get its item local path.
        :param add_to_future_clearing_path: Whether to add the local path to the future clearing paths list. If None, it
                                            will add the path to the list only if the data item is not of kind 'file',
                                            meaning it represents a local file and hence we don't want to delete it post
                                            running automatically. We wish to delete it only if the local path is
                                            temporary (and that will be in case kind is not 'file', so it is being
                                            downloaded to a temporary directory).

        :return: The data item local path.
        """
        # Get the local path to the item handled by the data item (download it to temporary if not local already):
        local_path = data_item.local()

        # Check if needed to add to the future clear list:
        if add_to_future_clearing_path or (
            add_to_future_clearing_path is None and data_item.kind != "file"
        ):
            self.add_future_clearing_path(path=local_path)

        return local_path



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

import mlrun
from mlrun.datastore.model_provider.model_provider import (
    InvokeResponseFormat,
    ModelProvider,
    UsageResponseKeys,
)


class MockModelProvider(ModelProvider):
    support_async = False

    def __init__(
        self,
        parent,
        kind,
        name,
        endpoint="",
        secrets: Optional[dict] = None,
        default_invoke_kwargs: Optional[dict] = None,
    ):
        super().__init__(
            parent=parent, name=name, kind=kind, endpoint=endpoint, secrets=secrets
        )
        self.default_invoke_kwargs = default_invoke_kwargs or {}
        self._client = None
        self._async_client = None

    @staticmethod
    def _extract_string_output(response: Any) -> str:
        """
        Extracts string response from response object
        """
        pass

    def load_client(self) -> None:
        """
        Initializes the SDK client for the model provider with the given keyword arguments
        and assigns it to an instance attribute (e.g., self._client).

        Subclasses should override this method to:
        - Create and configure the provider-specific client instance.
        - Assign the client instance to self._client.
        """

        pass

    def invoke(
        self,
        messages: Union[list[dict], Any],
        invoke_response_format: InvokeResponseFormat = InvokeResponseFormat.FULL,
        **invoke_kwargs,
    ) -> Union[str, dict[str, Any], Any]:
        if invoke_response_format == InvokeResponseFormat.STRING:
            return (
                "You are using a mock model provider, no actual inference is performed."
            )
        elif invoke_response_format == InvokeResponseFormat.FULL:
            return {
                UsageResponseKeys.USAGE: {"prompt_tokens": 0, "completion_tokens": 0},
                UsageResponseKeys.ANSWER: "You are using a mock model provider, no actual inference is performed.",
                "extra": {},
            }
        elif invoke_response_format == InvokeResponseFormat.USAGE:
            return {
                UsageResponseKeys.ANSWER: "You are using a mock model provider, no actual inference is performed.",
                UsageResponseKeys.USAGE: {"prompt_tokens": 0, "completion_tokens": 0},
            }
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(
                f"Unsupported invoke response format: {invoke_response_format}"
            )
